"""IAM utilities for building policy statements and granting permissions.

This module provides predefined IAM actions lists and helper functions for
creating policy statements for various AWS services.

Note:
    The list of actions for each service is incomplete and based on project needs.
    A helpful resource to research actions is: https://www.awsiamactions.io/
"""

from aibs_informatics_core.env import EnvBase
from aws_cdk import aws_iam as iam

from aibs_informatics_cdk_lib.common.aws.core_utils import (
    build_arn,
    build_batch_arn,
    build_dynamodb_arn,
    build_lambda_arn,
    build_s3_arn,
    build_sfn_arn,
)

#
# utils
#


def grant_managed_policies(
    role: iam.IRole | None,
    *managed_policies: str | iam.ManagedPolicy,
) -> None:
    """Grant managed policies to an IAM role.

    Args:
        role (Optional[iam.IRole]): The IAM role to grant policies to.
            If None, no action is taken.
        *managed_policies (Union[str, iam.ManagedPolicy]): Variable number of
            managed policies to grant. Can be policy names (str) or
            ManagedPolicy objects.
    """
    if not role:
        return

    for mp in managed_policies:
        role.add_managed_policy(
            iam.ManagedPolicy.from_aws_managed_policy_name(mp) if isinstance(mp, str) else mp
        )


#
# policy action lists
#

BATCH_READ_ONLY_ACTIONS = [
    "batch:Describe*",
    "batch:List*",
]

BATCH_FULL_ACCESS_ACTIONS = [
    "batch:RegisterJobDefinition",
    "batch:DeregisterJobDefinition",
    "batch:DescribeJobDefinitions",
    *BATCH_READ_ONLY_ACTIONS,
    "batch:*",
]

CLOUDWATCH_READ_ACTIONS = [
    "logs:DescribeLogGroups",
    "logs:GetLogEvents",
    "logs:GetLogGroupFields",
    "logs:GetLogRecord",
    "logs:GetQueryResults",
]

CLOUDWATCH_WRITE_ACTIONS = [
    "logs:CreateLogGroup",
    "logs:CreateLogStream",
    "logs:PutLogEvents",
]

CLOUDWATCH_FULL_ACCESS_ACTIONS = [
    *CLOUDWATCH_READ_ACTIONS,
    *CLOUDWATCH_WRITE_ACTIONS,
]

DYNAMODB_READ_ACTIONS = [
    "dynamodb:BatchGet*",
    "dynamodb:DescribeStream",
    "dynamodb:DescribeTable",
    "dynamodb:Get*",
    "dynamodb:Query",
    "dynamodb:Scan",
]

DYNAMODB_WRITE_ACTIONS = [
    "dynamodb:BatchWrite*",
    "dynamodb:CreateTable",
    "dynamodb:Delete*",
    "dynamodb:Update*",
    "dynamodb:PutItem",
]

DYNAMODB_READ_WRITE_ACTIONS = [
    *DYNAMODB_READ_ACTIONS,
    *DYNAMODB_WRITE_ACTIONS,
]


EC2_ACTIONS = ["ec2:DescribeAvailabilityZones"]

ECS_READ_ACTIONS = [
    "ecs:DescribeContainerInstances",
    "ecs:DescribeTaskDefinition",
    "ecs:DescribeTasks",
    "ecs:ListTasks",
]

ECS_WRITE_ACTIONS = [
    "ecs:RegisterTaskDefinition",
]

ECS_RUN_ACTIONS = [
    "ecs:RunTask",
]

ECS_FULL_ACCESS_ACTIONS = [
    *ECS_READ_ACTIONS,
    *ECS_WRITE_ACTIONS,
    *ECS_RUN_ACTIONS,
]

ECR_READ_ACTIONS = [
    "ecr:BatchCheckLayerAvailability",
    "ecr:BatchGetImage",
    "ecr:DescribeImageScanFindings",
    "ecr:DescribeImages",
    "ecr:DescribeRepositories",
    "ecr:GetAuthorizationToken",
    "ecr:GetDownloadUrlForLayer",
    "ecr:GetRepositoryPolicy",
    "ecr:ListImages",
    "ecr:ListTagsForResource",
]

ECR_WRITE_ACTIONS = [
    "ecr:CompleteLayerUpload",
    "ecr:CreateRepository",
    "ecr:DeleteRepository",
    "ecr:InitiateLayerUpload",
    "ecr:PutImage",
    "ecr:PutLifecyclePolicy",
    "ecr:UploadLayerPart",
]

ECR_TAGGING_ACTIONS = [
    "ecr:TagResource",
    "ecr:UntagResource",
]

ECR_FULL_ACCESS_ACTIONS = [
    *ECR_READ_ACTIONS,
    *ECR_TAGGING_ACTIONS,
    *ECR_WRITE_ACTIONS,
]

KMS_READ_ACTIONS = [
    "kms:Decrypt",
    "kms:DescribeKey",
]

KMS_WRITE_ACTIONS = [
    "kms:Encrypt",
    "kms:GenerateDataKey*",
    "kms:PutKeyPolicy",
]

KMS_FULL_ACCESS_ACTIONS = [
    *KMS_READ_ACTIONS,
    *KMS_WRITE_ACTIONS,
]


LAMBDA_FULL_ACCESS_ACTIONS = ["lambda:*"]
LAMBDA_READ_ONLY_ACTIONS = [
    "lambda:Get*",
    "lambda:List*",
]

S3_FULL_ACCESS_ACTIONS = ["s3:*"]

S3_READ_ONLY_ACCESS_ACTIONS = [
    "s3:Get*",
    "s3:List*",
    "s3-object-lambda:Get*",
    "s3-object-lambda:List*",
]

SECRETSMANAGER_READ_ONLY_ACTIONS = [
    "secretsmanager:DescribeSecret",
    "secretsmanager:GetRandomPassword",
    "secretsmanager:GetResourcePolicy",
    "secretsmanager:GetSecretValue",
    "secretsmanager:ListSecretVersionIds",
    "secretsmanager:ListSecrets",
]

SECRETSMANAGER_WRITE_ACTIONS = [
    "secretsmanager:CreateSecret",
    "secretsmanager:PutSecretValue",
    "secretsmanager:ReplicateSecretToRegions",
    "secretsmanager:RestoreSecret",
    "secretsmanager:RotateSecret",
    "secretsmanager:UpdateSecret",
    "secretsmanager:UpdateSecretVersionStage",
]

SECRETSMANAGER_DELETE_ACTIONS = [
    "secretsmanager:CancelRotateSecret",
    "secretsmanager:DeleteSecret",
    "secretsmanager:RemoveRegionsFromReplication",
    "secretsmanager:StopReplicationToReplica",
]

SECRETSMANAGER_READ_WRITE_ACTIONS = [
    *SECRETSMANAGER_READ_ONLY_ACTIONS,
    *SECRETSMANAGER_WRITE_ACTIONS,
]

SECRETSMANAGER_FULL_ADMIN_ACTIONS = [
    *SECRETSMANAGER_READ_ONLY_ACTIONS,
    *SECRETSMANAGER_WRITE_ACTIONS,
    *SECRETSMANAGER_DELETE_ACTIONS,
]


SES_FULL_ACCESS_ACTIONS = ["ses:*"]

SFN_STATES_READ_ACCESS_ACTIONS = [
    "states:DescribeActivity",
    "states:DescribeExecution",
    "states:DescribeStateMachine",
    "states:DescribeStateMachineForExecution",
    "states:ListExecutions",
    "states:GetExecutionHistory",
    "states:ListStateMachines",
    "states:ListActivities",
]

SFN_STATES_EXECUTION_ACTIONS = [
    "states:StartExecution",
    "states:StopExecution",
]

SNS_FULL_ACCESS_ACTIONS = ["sns:*"]

SQS_READ_ACTIONS = [
    "sqs:GetQueueAttributes",
    "sqs:GetQueueUrl",
    "sqs:ReceiveMessage",
    "sqs:SendMessage",
]

SQS_WRITE_ACTIONS = [
    "sqs:ChangeMessageVisibility",
    "sqs:DeleteMessage",
]

SQS_FULL_ACCESS_ACTIONS = [
    *SQS_READ_ACTIONS,
    *SQS_WRITE_ACTIONS,
]

SSM_READ_ACTIONS = [
    "ssm:GetParameter",
    "ssm:GetParameters",
    "ssm:GetParametersByPath",
]


#
# policy statement constants and builders
#

CODE_BUILD_IAM_POLICY = iam.PolicyStatement(
    actions=[
        *EC2_ACTIONS,
        *ECR_FULL_ACCESS_ACTIONS,
    ],
    resources=["*"],
)


def batch_policy_statement(
    env_base: EnvBase | None = None,
    actions: list[str] = BATCH_FULL_ACCESS_ACTIONS,
    sid: str = "BatchReadWrite",
) -> iam.PolicyStatement:
    """Create an IAM policy statement for AWS Batch.

    Args:
        env_base (Optional[EnvBase]): Environment base for resource prefix.
            Defaults to None (matches all).
        actions (List[str]): List of Batch actions to allow.
            Defaults to BATCH_FULL_ACCESS_ACTIONS.
        sid (str): Statement ID. Defaults to "BatchReadWrite".

    Returns:
        IAM policy statement for Batch resources.
    """
    resource_id = f"{env_base or ''}*"

    return iam.PolicyStatement(
        sid=sid,
        actions=actions,
        effect=iam.Effect.ALLOW,
        resources=[
            build_batch_arn(
                resource_id=resource_id,
                resource_type="compute-environment",
            ),
            build_batch_arn(
                resource_id=resource_id,
                resource_type="job",
            ),
            build_batch_arn(
                resource_id=resource_id,
                resource_type="job-definition",
            ),
            build_batch_arn(
                resource_id=resource_id,
                resource_type="job-queue",
            ),
            # ERROR: An error occurred (AccessDeniedException) when calling the
            # DescribeJobDefinitions operation:
            # User: arn:aws:sts::051791135335:assumed-role/Infrastructure.../dev-ryan-gwo-create-job-definition-fn  # noqa: E501
            # is not authorized to perform: batch:DescribeJobDefinitions on resource: "*"
            # TODO: WTF why does this not work... adding "*" resource for now
            "*",
        ],
    )


def dynamodb_policy_statement(
    env_base: EnvBase | None = None,
    actions: list[str] = DYNAMODB_READ_WRITE_ACTIONS,
    sid: str = "DynamoDBReadWrite",
) -> iam.PolicyStatement:
    """Create an IAM policy statement for DynamoDB.

    Args:
        env_base (Optional[EnvBase]): Environment base for resource prefix.
            Defaults to None (matches all).
        actions (List[str]): List of DynamoDB actions to allow.
            Defaults to DYNAMODB_READ_WRITE_ACTIONS.
        sid (str): Statement ID. Defaults to "DynamoDBReadWrite".

    Returns:
        IAM policy statement for DynamoDB tables.
    """
    return iam.PolicyStatement(
        sid=sid,
        actions=actions,
        effect=iam.Effect.ALLOW,
        resources=[
            build_dynamodb_arn(
                resource_id=f"{env_base or ''}*",
                resource_type="table",
            ),
        ],
    )


def ecs_policy_statement(
    actions: list[str] = ECS_READ_ACTIONS, sid: str = "ECSDescribe"
) -> iam.PolicyStatement:
    """Create an IAM policy statement for ECS.

    Args:
        actions (List[str]): List of ECS actions to allow.
            Defaults to ECS_READ_ACTIONS.
        sid (str): Statement ID. Defaults to "ECSDescribe".

    Returns:
        IAM policy statement for ECS resources.
    """
    return iam.PolicyStatement(
        sid=sid,
        actions=actions,
        effect=iam.Effect.ALLOW,
        resources=[
            build_arn(
                service="ecs",
                resource_id="*/*",
                resource_type="container-instance",
                resource_delim="/",
            ),
        ],
    )


def lambda_policy_statement(
    env_base: EnvBase | None = None,
    actions: list[str] = LAMBDA_FULL_ACCESS_ACTIONS,
    sid: str = "LambdaReadWrite",
) -> iam.PolicyStatement:
    """Create an IAM policy statement for Lambda.

    Args:
        env_base (Optional[EnvBase]): Environment base for resource prefix.
            Defaults to None (matches all).
        actions (List[str]): List of Lambda actions to allow.
            Defaults to LAMBDA_FULL_ACCESS_ACTIONS.
        sid (str): Statement ID. Defaults to "LambdaReadWrite".

    Returns:
        IAM policy statement for Lambda functions.
    """
    return iam.PolicyStatement(
        sid=sid,
        actions=actions,
        effect=iam.Effect.ALLOW,
        resources=[
            build_lambda_arn(
                resource_id=f"{env_base or ''}*",
                resource_type="function",
            ),
        ],
    )


def s3_policy_statement(
    env_base: EnvBase | None = None,
    actions: list[str] = S3_FULL_ACCESS_ACTIONS,
    sid: str = "S3FullAccess",
) -> iam.PolicyStatement:
    """Create an IAM policy statement for S3.

    Args:
        env_base (Optional[EnvBase]): Environment base for resource prefix.
            Defaults to None (matches all).
        actions (List[str]): List of S3 actions to allow.
            Defaults to S3_FULL_ACCESS_ACTIONS.
        sid (str): Statement ID. Defaults to "S3FullAccess".

    Returns:
        IAM policy statement for S3 buckets.
    """
    return iam.PolicyStatement(
        sid=sid,
        actions=actions,
        effect=iam.Effect.ALLOW,
        resources=[
            build_s3_arn(
                resource_id=f"{env_base or ''}*",
                resource_type="bucket",
            ),
        ],
    )


def secretsmanager_policy_statement(
    actions: list[str] = SECRETSMANAGER_READ_ONLY_ACTIONS,
    sid: str = "SecretsManagerReadOnly",
    resource_id: str = "*",
    region: str = None,
    account: str = None,
) -> iam.PolicyStatement:
    """Create an IAM policy statement for Secrets Manager.

    Args:
        actions (List[str]): List of Secrets Manager actions to allow.
            Defaults to SECRETSMANAGER_READ_ONLY_ACTIONS.
        sid (str): Statement ID. Defaults to "SecretsManagerReadOnly".
        resource_id (str): Resource identifier. Defaults to "*".
        region (str): AWS region. Defaults to None (current region).
        account (str): AWS account ID. Defaults to None (current account).

    Returns:
        IAM policy statement for Secrets Manager resources.
    """
    return iam.PolicyStatement(
        sid=sid,
        actions=actions,
        effect=iam.Effect.ALLOW,
        resources=[
            build_arn(
                service="secretsmanager",
                resource_id=resource_id,
                region=region,
                account=account,
            ),
        ],
    )


def ses_policy_statement(
    actions: list[str] = SES_FULL_ACCESS_ACTIONS,
    sid: str = "SESFullAccess",
) -> iam.PolicyStatement:
    """Create an IAM policy statement for SES.

    Args:
        actions (List[str]): List of SES actions to allow.
            Defaults to SES_FULL_ACCESS_ACTIONS.
        sid (str): Statement ID. Defaults to "SESFullAccess".

    Returns:
        IAM policy statement for SES resources.
    """
    return iam.PolicyStatement(
        sid=sid,
        actions=actions,
        effect=iam.Effect.ALLOW,
        resources=[
            build_arn(
                service="ses",
            ),
        ],
    )


def sfn_policy_statement(
    env_base: EnvBase | None = None,
    actions: list[str] = SFN_STATES_READ_ACCESS_ACTIONS,
    sid: str = "SfnFullAccess",
) -> iam.PolicyStatement:
    """Create an IAM policy statement for Step Functions.

    Args:
        env_base (Optional[EnvBase]): Environment base for resource prefix.
            Defaults to None (matches all).
        actions (List[str]): List of Step Functions actions to allow.
            Defaults to SFN_STATES_READ_ACCESS_ACTIONS.
        sid (str): Statement ID. Defaults to "SfnFullAccess".

    Returns:
        IAM policy statement for Step Functions resources.
    """
    return iam.PolicyStatement(
        sid=sid,
        actions=actions,
        effect=iam.Effect.ALLOW,
        resources=[
            build_sfn_arn(
                resource_id=f"{env_base or ''}*",
                resource_type="*",
            ),
        ],
    )


def sns_policy_statement(
    actions: list[str] = SNS_FULL_ACCESS_ACTIONS,
    sid: str = "SNSFullAccess",
) -> iam.PolicyStatement:
    """Create an IAM policy statement for SNS.

    Args:
        actions (List[str]): List of SNS actions to allow.
            Defaults to SNS_FULL_ACCESS_ACTIONS.
        sid (str): Statement ID. Defaults to "SNSFullAccess".

    Returns:
        IAM policy statement for SNS resources.
    """
    return iam.PolicyStatement(
        sid=sid,
        actions=actions,
        effect=iam.Effect.ALLOW,
        resources=[
            build_arn(
                service="sns",
            ),
        ],
    )


def ssm_policy_statement(
    actions: list[str] = SSM_READ_ACTIONS, sid: str = "SSMParamReadActions"
) -> iam.PolicyStatement:
    """Create an IAM policy statement for SSM Parameter Store.

    Args:
        actions (List[str]): List of SSM actions to allow.
            Defaults to SSM_READ_ACTIONS.
        sid (str): Statement ID. Defaults to "SSMParamReadActions".

    Returns:
        IAM policy statement for SSM resources.
    """
    return iam.PolicyStatement(
        sid=sid, actions=actions, effect=iam.Effect.ALLOW, resources=[build_arn(service="ssm")]
    )


def sqs_policy_statement(
    env_base: EnvBase | None = None,
    actions: list[str] = SQS_FULL_ACCESS_ACTIONS,
    sid: str = "SQSFullAccess",
) -> iam.PolicyStatement:
    """Create an IAM policy statement for SQS.

    Args:
        env_base (Optional[EnvBase]): Environment base for resource prefix.
            Defaults to None (matches all).
        actions (List[str]): List of SQS actions to allow.
            Defaults to SQS_FULL_ACCESS_ACTIONS.
        sid (str): Statement ID. Defaults to "SQSFullAccess".

    Returns:
        IAM policy statement for SQS queues.
    """
    return iam.PolicyStatement(
        sid=sid,
        actions=actions,
        effect=iam.Effect.ALLOW,
        resources=[
            build_arn(
                service="sqs",
                resource_id=f"{env_base or ''}*",
            )
        ],
    )



from typing import Any, Literal

import constructs
from aws_cdk import aws_stepfunctions as sfn

from aibs_informatics_cdk_lib.common.aws.sfn_utils import JsonReferencePath


class CommonOperation:
    @classmethod
    def merge_defaults(
        cls,
        scope: constructs.Construct,
        id: str,
        defaults: dict[str, Any],
        input_path: str = "$",
        target_path: str = "$",
        result_path: str | None = None,
        order_of_preference: Literal["target", "default"] = "target",
        check_if_target_present: bool = False,
    ) -> sfn.Chain:
        """Wrapper chain that merges input with defaults.

        Notes:
            - reference paths in defaults should be relative to the input path

        Args:
            scope (constructs.Construct): construct scope
            id (str): identifier for the states created
            defaults (dict[str, Any]): default values to merge with input. If any reference paths
                are present in the defaults, they should be relative to the input path.
            input_path (str, optional): Input path of object to merge. De faults to "$".
            target_path (str, optional): target path to merge with. This should be relative to
                the input_path parameter. Defaults to "$".
            result_path (Optional[str], optional): result path to store merged results.
                If not specified, it defaults to the target_path relative to input_path.
                If specified, it is considered an absolute path.
            order_of_preference (Literal["target", "default"], optional): If "target", the target
                path will be merged with the defaults. If "default", the defaults will be merged
                with the target path. Defaults to "target".
            check_if_target_present (bool, optional): If true, check if the target path is present
                in the input. If not, the defaults will be used as the result. Otherwise, the
                defaults will be merged with the target path. This is useful for optional
                parameters that may or may not be present in the input.
                Defaults to False.

        Returns:
            sfn.Chain: the new chain that merges defaults with input
        """
        input_path = JsonReferencePath(input_path)
        target_path = JsonReferencePath(target_path)
        result_path = (
            JsonReferencePath(result_path) if result_path else input_path.extend(target_path)
        )

        pref1 = "$.target" if order_of_preference == "target" else "$.default"
        pref2 = "$.default" if order_of_preference == "target" else "$.target"

        merge_task = sfn.Pass(
            scope,
            "Merge Pass",
            parameters={
                "merged": sfn.JsonPath.json_merge(
                    sfn.JsonPath.object_at(pref2),
                    sfn.JsonPath.object_at(pref1),
                ),
            },
            output_path="$.merged",
        )
        chain_start: sfn.Pass | sfn.Choice
        if check_if_target_present:
            # Branch based on presence of the target
            choice = sfn.Choice(scope, "Check Target")
            present_pass = sfn.Pass(
                scope,
                "Target Present",
                parameters={
                    "target": sfn.JsonPath.object_at(target_path.as_reference),
                    "default": defaults,
                },
            )
            not_present_pass = sfn.Pass(
                scope,
                "Target Not Present",
                parameters={
                    "target": {},
                    "default": defaults,
                },
            )
            # Chain both branches into the merge task
            present_pass.next(merge_task)
            not_present_pass.next(merge_task)
            choice.when(
                sfn.Condition.is_present(target_path.as_reference),
                present_pass,
            ).otherwise(not_present_pass)
            chain_start = choice
        else:
            init_pass = sfn.Pass(
                scope,
                "Init Pass",
                parameters={
                    "target": sfn.JsonPath.object_at(target_path.as_reference),
                    "default": defaults,
                },
            )
            init_pass.next(merge_task)
            chain_start = init_pass

        parallel = sfn.Chain.start(chain_start).to_single_state(
            id=id,
            input_path=input_path.as_reference,
            result_path=result_path.as_reference,
        )
        restructure = sfn.Pass(
            scope,
            f"{id} Restructure",
            input_path=f"{result_path.as_reference}[0]",
            result_path=result_path.as_reference,
        )
        return parallel.next(restructure)

    @classmethod
    def enclose_chainable(
        cls,
        scope: constructs.Construct,
        id: str,
        definition: sfn.IChainable,
        input_path: str | None = None,
        result_path: str | None = None,
    ) -> sfn.Chain:
        """Enclose the current state machine fragment within a parallel state.

        Notes:
            - If input_path is not provided, it will default to "$"
            - If result_path is not provided, it will default to input_path

        Args:
            id (str): an identifier for the parallel state
            input_path (Optional[str], optional): input path for the enclosed state.
                Defaults to "$".
            result_path (Optional[str], optional): result path to put output of enclosed state.
                Defaults to same as input_path.

        Returns:
            sfn.Chain: the new state machine fragment
        """
        if input_path is None:
            input_path = "$"
        if result_path is None:
            result_path = input_path

        chain = (
            sfn.Chain.start(definition)
            if not isinstance(definition, (sfn.Chain, sfn.StateMachineFragment))
            else definition
        )

        if isinstance(chain, sfn.Chain):
            parallel = chain.to_single_state(
                id=f"{id} Enclosure", input_path=input_path, result_path=result_path
            )
        else:
            parallel = chain.to_single_state(input_path=input_path, result_path=result_path)
        definition = sfn.Chain.start(parallel)

        if result_path and result_path != sfn.JsonPath.DISCARD:
            restructure = sfn.Pass(
                scope,
                f"{id} Enclosure Post",
                input_path=f"{result_path}[0]",
                result_path=result_path,
            )
            definition = definition.next(restructure)

        return definition



from collections.abc import Iterable

import constructs
from aibs_informatics_core.env import EnvBase
from aws_cdk import aws_batch as batch
from aws_cdk import aws_ecr_assets as ecr_assets
from aws_cdk import aws_iam as iam
from aws_cdk import aws_s3 as s3
from aws_cdk import aws_stepfunctions as sfn

from aibs_informatics_cdk_lib.common.aws.iam_utils import (
    SFN_STATES_EXECUTION_ACTIONS,
    SFN_STATES_READ_ACCESS_ACTIONS,
    sfn_policy_statement,
)
from aibs_informatics_cdk_lib.constructs_.base import EnvBaseConstructMixins
from aibs_informatics_cdk_lib.constructs_.efs.file_system import MountPointConfiguration
from aibs_informatics_cdk_lib.constructs_.sfn.fragments.informatics.batch import (
    BatchInvokedBaseFragment,
    BatchInvokedLambdaFunction,
)


class DataSyncFragment(BatchInvokedBaseFragment, EnvBaseConstructMixins):
    def __init__(
        self,
        scope: constructs.Construct,
        id: str,
        env_base: EnvBase,
        aibs_informatics_docker_asset: ecr_assets.DockerImageAsset | str,
        batch_job_queue: batch.JobQueue | str,
        scaffolding_bucket: s3.Bucket,
        batch_job_role: iam.Role | str | None = None,
        mount_point_configs: Iterable[MountPointConfiguration] | None = None,
    ) -> None:
        """Sync data from one s3 bucket to another


        Args:
            scope (Construct): construct scope
            id (str): id
            env_base (EnvBase): env base
            aibs_informatics_docker_asset (DockerImageAsset|str): Docker image asset or image uri
                str for the aibs informatics aws lambda
            batch_job_queue (JobQueue|str): Default batch job queue or job queue name str that
                the batch job will be submitted to. This can be override by the payload.
            scaffolding_bucket (Bucket): Primary bucket used for request/response json blobs used
                in the batch invoked lambda function.
            batch_job_role (Optional[IRole|str], optional): Optional role to use for the batch job.
                If not provided, the default role created by the batch compute construct will be
                used.
            mount_point_configs (Optional[Iterable[MountPointConfiguration]], optional):
                List of mount point configurations to use. These can be overridden in the payload.

        """
        super().__init__(scope, id, env_base)

        aibs_informatics_image_uri = (
            aibs_informatics_docker_asset
            if isinstance(aibs_informatics_docker_asset, str)
            else aibs_informatics_docker_asset.image_uri
        )

        self.batch_job_queue_name = (
            batch_job_queue if isinstance(batch_job_queue, str) else batch_job_queue.job_queue_name
        )

        start = sfn.Pass(
            self,
            "Input Restructure",
            parameters={
                "handler": "aibs_informatics_aws_lambda.handlers.data_sync.data_sync_handler",
                "image": aibs_informatics_image_uri,
                "payload": sfn.JsonPath.object_at("$"),
            },
        )

        self.fragment = BatchInvokedLambdaFunction.with_defaults(
            self,
            "Data Sync",
            env_base=self.env_base,
            name="data-sync",
            job_queue=self.batch_job_queue_name,
            bucket_name=scaffolding_bucket.bucket_name,
            handler_path="$.handler",
            image_path="$.image",
            payload_path="$.payload",
            memory="1024",
            vcpus="1",
            mount_point_configs=list(mount_point_configs) if mount_point_configs else None,
            job_role_arn=(
                batch_job_role if isinstance(batch_job_role, str) else batch_job_role.role_arn
            )
            if batch_job_role
            else None,
            environment={
                EnvBase.ENV_BASE_KEY: self.env_base,
                "AWS_REGION": self.aws_region,
                "AWS_ACCOUNT_ID": self.aws_account,
            },
        )

        self.definition = start.next(self.fragment.to_single_state())

    @property
    def required_managed_policies(self) -> list[iam.IManagedPolicy | str]:
        return [
            *super().required_managed_policies,
            *[_ for _ in self.fragment.required_managed_policies],
        ]

    @property
    def required_inline_policy_statements(self) -> list[iam.PolicyStatement]:
        return [
            *self.fragment.required_inline_policy_statements,
            *super().required_inline_policy_statements,
            sfn_policy_statement(
                self.env_base,
                actions=SFN_STATES_EXECUTION_ACTIONS + SFN_STATES_READ_ACCESS_ACTIONS,
            ),
        ]


class DistributedDataSyncFragment(BatchInvokedBaseFragment):
    def __init__(
        self,
        scope: constructs.Construct,
        id: str,
        env_base: EnvBase,
        aibs_informatics_docker_asset: ecr_assets.DockerImageAsset | str,
        batch_job_queue: batch.JobQueue | str,
        scaffolding_bucket: s3.Bucket,
        batch_job_role: str | iam.Role | None = None,
        mount_point_configs: Iterable[MountPointConfiguration] | None = None,
    ) -> None:
        """Sync data from one s3 bucket to another using distributed batch jobs

        Args:
            scope (constructs.Construct): construct scope
            id (str): id
            env_base (EnvBase): env base
            aibs_informatics_docker_asset (DockerImageAsset|str): Docker image asset or image uri
                str for the aibs informatics aws lambda
            batch_job_queue (JobQueue|str): Default batch job queue or job queue name str that
                the batch job will be submitted to. This can be override by the payload.
            scaffolding_bucket (Bucket): Primary bucket used for request/response json blobs used
                in the batch invoked lambda function.
            batch_job_role (Optional[IRole|str], optional): Optional role to use for the batch job.
                If not provided, the default role created by the batch compute construct will be
                used.
            mount_point_configs (Optional[Iterable[MountPointConfiguration]], optional):
                List of mount point configurations to use. These can be overridden in the payload.
        """
        super().__init__(scope, id, env_base)
        start_pass_state = sfn.Pass(
            self,
            f"{id}: Start",
            parameters={
                "request": sfn.JsonPath.object_at("$"),
            },
        )
        prep_batch_sync_task_name = "prep-batch-data-sync-requests"

        prep_batch_sync = BatchInvokedLambdaFunction(
            scope=scope,
            id=f"{id}: Prep Batch Data Sync",
            env_base=env_base,
            name=prep_batch_sync_task_name,
            payload_path="$.request",
            image=(
                aibs_informatics_docker_asset
                if isinstance(aibs_informatics_docker_asset, str)
                else aibs_informatics_docker_asset.image_uri
            ),
            handler="aibs_informatics_aws_lambda.handlers.data_sync.prepare_batch_data_sync_handler",
            job_queue=(
                batch_job_queue
                if isinstance(batch_job_queue, str)
                else batch_job_queue.job_queue_name
            ),
            bucket_name=scaffolding_bucket.bucket_name,
            memory=1024,
            vcpus=1,
            mount_point_configs=list(mount_point_configs) if mount_point_configs else None,
            job_role_arn=(
                batch_job_role if isinstance(batch_job_role, str) else batch_job_role.role_arn
            )
            if batch_job_role
            else None,
        ).enclose(result_path=f"$.tasks.{prep_batch_sync_task_name}.response")

        batch_sync_map_state = sfn.Map(
            self,
            f"{id}: Batch Data Sync: Map Start",
            comment="Runs requests for batch sync in parallel",
            items_path=f"$.tasks.{prep_batch_sync_task_name}.response.requests",
            result_path=sfn.JsonPath.DISCARD,
        )

        batch_sync_map_state.iterator(
            BatchInvokedLambdaFunction(
                scope=scope,
                id=f"{id}: Batch Data Sync",
                env_base=env_base,
                name="batch-data-sync",
                payload_path="$",
                image=(
                    aibs_informatics_docker_asset
                    if isinstance(aibs_informatics_docker_asset, str)
                    else aibs_informatics_docker_asset.image_uri
                ),
                handler="aibs_informatics_aws_lambda.handlers.data_sync.batch_data_sync_handler",
                job_queue=(
                    batch_job_queue
                    if isinstance(batch_job_queue, str)
                    else batch_job_queue.job_queue_name
                ),
                bucket_name=scaffolding_bucket.bucket_name,
                memory=4096,
                vcpus=2,
                mount_point_configs=list(mount_point_configs) if mount_point_configs else None,
                job_role_arn=(
                    batch_job_role if isinstance(batch_job_role, str) else batch_job_role.role_arn
                )
                if batch_job_role
                else None,
            )
        )
        # fmt: off
        self.definition = (
            start_pass_state
            .next(prep_batch_sync)
            .next(batch_sync_map_state)
        )
        # fmt: on



"""Step Functions utilities for JSON path references.

This module provides utilities for working with Step Functions JSON path expressions.
"""

import re
from functools import reduce
from re import Pattern
from typing import Any, ClassVar, cast

import aws_cdk as cdk
from aws_cdk import aws_stepfunctions as sfn


class JsonReferencePath(str):
    """String extension for defining JsonPath reference expressions.

    Provides properties and methods for working with JSON path references
    in AWS Step Functions state machines.

    More details: https://github.com/json-path/JsonPath

    Primarily supports "$" reference.

    Example:
        >>> path = JsonReferencePath("input.data")
        >>> path.as_reference
        '$.input.data'
        >>> path.as_key
        'input.data.$'
    """

    _EXTRA_PERIODS_PATTERN: ClassVar[Pattern[str]] = re.compile(r"[$.]+")
    _PERIOD_PATTERN: ClassVar[Pattern[str]] = re.compile(r"(?<!\$)\.(?!\$)")
    _PREFIX: ClassVar[str] = "$."
    _SUFFIX: ClassVar[str] = ".$"
    _REF: ClassVar[str] = "$"

    def __new__(cls, content: str):
        """Create a new JsonReferencePath.

        Args:
            content (str): The path content to wrap.

        Returns:
            A sanitized JsonReferencePath instance.
        """
        return super().__new__(cls, cls.sanitize(content))

    def __add__(self, other):
        """Concatenate paths with a period separator.

        Args:
            other: The path segment to append.

        Returns:
            A new JsonReferencePath with the appended segment.
        """
        return JsonReferencePath(super().__add__("." + other))

    def extend(self, *paths: str) -> "JsonReferencePath":
        """Extend the path with multiple segments.

        Args:
            *paths (str): Variable number of path segments to append.

        Returns:
            A new JsonReferencePath with all segments appended.
        """
        return cast(JsonReferencePath, reduce(lambda x, y: x + y, [self, *paths]))

    @property
    def as_key(self) -> str:
        """Return the reference path as a key.

        Appends ".$" suffix for use as a state machine key.

        Returns:
            The path formatted as a key.
        """
        return f"{self}{self._SUFFIX}" if self else self._REF

    @property
    def as_reference(self) -> str:
        """Return the reference path as a value.

        Prepends "$." prefix for use as a state machine reference.

        Returns:
            The path formatted as a reference.
        """
        return f"{self._PREFIX}{self}" if self else self._REF

    @property
    def as_jsonpath_string(self) -> str:
        """Return the path as a Step Functions string reference.

        Returns:
            The path wrapped in JsonPath.string_at().
        """
        return sfn.JsonPath.string_at(self.as_reference)

    @property
    def as_jsonpath_object(self) -> cdk.IResolvable:
        """Return the path as a Step Functions object reference.

        Returns:
            The path wrapped in JsonPath.object_at().
        """
        return sfn.JsonPath.object_at(self.as_reference)

    @property
    def as_jsonpath_json_to_string(self) -> str:
        """Return the path as a JSON-to-string conversion.

        Returns:
            The object reference converted to string via JsonPath.json_to_string().
        """
        return sfn.JsonPath.json_to_string(self.as_jsonpath_object)

    @property
    def as_jsonpath_list(self) -> list[str]:
        """Return the path as a Step Functions list reference.

        Returns:
            The path wrapped in JsonPath.list_at().
        """
        return sfn.JsonPath.list_at(self.as_reference)

    @property
    def as_jsonpath_number(self) -> int | float:
        """Return the path as a Step Functions number reference.

        Returns:
            The path wrapped in JsonPath.number_at().
        """
        return sfn.JsonPath.number_at(self.as_reference)

    @classmethod
    def sanitize(cls, s: str) -> str:
        """Sanitize a string for use as a JSON path.

        Ensures string has non-consecutive periods and no periods at edges.

        Args:
            s (str): The string to sanitize.

        Returns:
            The sanitized string.
        """
        return f"{cls._EXTRA_PERIODS_PATTERN.sub('.', s).strip('.')}"

    @classmethod
    def is_reference(cls, s: Any) -> bool:
        """Check if a value is a JSON path reference.

        Args:
            s (Any): The value to check.

        Returns:
            True if the value is a JsonReferencePath or starts with "$".
        """
        return isinstance(s, JsonReferencePath) or isinstance(s, str) and s.startswith("$")

    @classmethod
    def empty(cls) -> "JsonReferencePath":
        """Create an empty JsonReferencePath.

        Returns:
            An empty JsonReferencePath instance.
        """
        return cls("")



from dataclasses import dataclass
from typing import Protocol

from aibs_informatics_core.env import EnvBase


class IBatchEnvironmentDescriptor(Protocol):
    def get_job_queue_name(self, env_base: EnvBase) -> str: ...

    def get_compute_environment_name(self, env_base: EnvBase) -> str: ...

    def get_name(self) -> str: ...


@dataclass
class BatchEnvironmentDescriptor:
    name: str

    def get_job_queue_name(self, env_base: EnvBase) -> str:
        return env_base.get_resource_name(self.name, "job-queue")

    def get_compute_environment_name(self, env_base: EnvBase) -> str:
        return env_base.get_resource_name(self.name, "ce")

    def get_name(self) -> str:
        return self.name



"""Git utilities for repository operations.

This module provides functions for working with Git repositories,
including URL parsing, cloning, and commit hash retrieval.
"""

import logging
import os
import re
import subprocess
import tempfile
from pathlib import Path
from typing import ClassVar

from aibs_informatics_core.collections import ValidatedStr
from aibs_informatics_core.utils.file_operations import remove_path

logger = logging.getLogger(__name__)


class GitUrl(ValidatedStr):
    """Validated string representing a Git repository URL.

    Supports various Git URL formats including HTTPS, SSH, and git protocols.
    Can extract repository name and optional ref (branch/tag/commit).

    Attributes:
        regex_pattern: Compiled regex pattern for URL validation.
    """

    regex_pattern: ClassVar[re.Pattern] = re.compile(
        r"((?:(?:git|ssh|http(?:s)?)(?::\/\/)(?:[\w\.]+)\/|(?:git@(?:[\w\.]+)):)(?:[\w\.-]+)\/(?:[\w\.-]+)(?:\.git)?)(?:(?:\#|@|\/tree\/)([\w\./-]+))?"
    )

    @property
    def repo_base_url(self) -> str:
        return f"{self.get_match_groups()[0].removesuffix('.git')}.git"

    @property
    def repo_name(self) -> str:
        return os.path.basename(self.repo_base_url.removesuffix(".git"))

    @property
    def ref(self) -> str | None:
        return self.get_match_groups()[-1]


def is_repo_url(url: str) -> bool:
    """Check if a URL is a valid Git repository URL.

    Args:
        url (str): The URL to validate.

    Returns:
        True if the URL is a valid Git repository URL, False otherwise.
    """
    return GitUrl.is_valid(url)


def is_local_repo(repo_path: str | Path) -> bool:
    """Check if a path is a local Git repository.

    Args:
        repo_path (Union[str, Path]): The file system path to check.

    Returns:
        True if the path is a local Git repository, False otherwise.
    """
    repo_path = Path(repo_path)
    try:
        subprocess.check_output(["git", "-C", repo_path, "rev-parse", "--is-inside-work-tree"])
        return True
    except subprocess.CalledProcessError:
        return False


def get_commit_hash(repo_url_or_path: str | Path) -> str | None:
    """Get the HEAD commit hash of a Git repository.

    Args:
        repo_url_or_path (Union[str, Path]): The repository URL or local path.

    Returns:
        The commit hash of the HEAD reference.

    Raises:
        ValueError: If the input is neither a valid URL nor a local repository.
    """
    if isinstance(repo_url_or_path, str) and is_repo_url(repo_url_or_path):
        return get_commit_hash_from_url(repo_url_or_path)
    elif is_local_repo(repo_url_or_path):
        repo_path = Path(repo_url_or_path)
        return get_commit_hash_from_local(repo_path)
    else:
        raise ValueError("The input must be a string or a Path object.")


def get_repo_url_components(repo_url: str) -> tuple[str, str | None]:
    """Extract base URL and ref from a Git repository URL.

    Args:
        repo_url (str): The full repository URL.

    Returns:
        Tuple of (base_url, ref) where ref may be None.
    """
    git_url = GitUrl(repo_url)
    return (git_url.repo_base_url, git_url.ref)


def get_commit_hash_from_url(repo_url: str) -> str:
    """Get the commit hash from a remote Git repository URL.

    Uses git ls-remote to fetch the commit hash without cloning.

    Args:
        repo_url (str): The repository URL.

    Returns:
        The commit hash of the HEAD or specified branch.

    Raises:
        subprocess.CalledProcessError: If git ls-remote fails.
    """
    try:
        url = GitUrl(repo_url)
        branch = url.ref or "HEAD"
        # Use git ls-remote to get the commit hashes of remote heads
        output = (
            subprocess.check_output(["git", "ls-remote", url.repo_base_url, "--branch", branch])
            .decode("utf-8")
            .strip()
        )
        # The first part of the output is the commit hash of the HEAD reference
        commit_hash = output.split("\t")[0]
        return commit_hash
    except subprocess.CalledProcessError as e:
        logger.error(f"An error occurred: {e}")
        raise e


def get_commit_hash_from_local(repo_path: str | Path) -> str:
    """Get the HEAD commit hash from a local Git repository.

    Args:
        repo_path (Union[str, Path]): Path to the local repository.

    Returns:
        The commit hash of HEAD.

    Raises:
        subprocess.CalledProcessError: If git rev-parse fails.
    """
    try:
        # Get the latest commit hash
        commit_hash = (
            subprocess.check_output(["git", "rev-parse", "HEAD"], cwd=repo_path)
            .decode("utf-8")
            .strip()
        )
        return commit_hash
    except subprocess.CalledProcessError as e:
        logger.error(
            f"An error occurred while trying to get the commit hash from local path {repo_path}: {e}"  # noqa: E501
        )
        raise e
    except Exception as e:
        logger.error(
            "An unexpected error occurred while trying to get the commit hash from local path "
            f"{repo_path}: {e}"
        )
        raise e


def get_repo_name(repo_url_or_path: str | Path) -> str:
    """Get the repository name from a URL or local path.

    Args:
        repo_url_or_path (Union[str, Path]): The repository URL or local path.

    Returns:
        The repository name.

    Raises:
        ValueError: If the input is neither a valid URL nor a local repository.
        subprocess.CalledProcessError: If git commands fail.
    """
    if isinstance(repo_url_or_path, str) and is_repo_url(repo_url_or_path):
        return GitUrl(repo_url_or_path).repo_name
    elif is_local_repo(repo_url_or_path):
        repo_path = Path(repo_url_or_path)
        try:
            # Get the remote URL of the 'origin' remote (commonly used name for the default remote)
            remote_url = (
                subprocess.check_output(
                    ["git", "config", "--get", "remote.origin.url"], cwd=repo_path
                )
                .decode("utf-8")
                .strip()
            )

            # Strip trailing slashes or .git if present
            remote_url = remote_url.rstrip("/").rstrip(".git")

            # Extract the repository name
            repo_name = os.path.basename(remote_url)

            return repo_name
        except subprocess.CalledProcessError as e:
            logger.error(f"An error occurred: {e}")
            raise e

    else:
        raise ValueError("The input must be a string or a Path object.")


def construct_repo_path(repo_url: str, target_dir: str | Path | None = None) -> Path:
    """Construct a deterministic path for a cloned repository.

    The path includes the repository name and commit hash to ensure
    unique paths for different versions.

    Args:
        repo_url (str): The repository URL.
        target_dir (Optional[Union[str, Path]]): Base directory for the path.
            Defaults to system temp directory.

    Returns:
        Path where the repository should be cloned.
    """
    target_dir = Path(target_dir) if target_dir else Path(tempfile.gettempdir())

    repo_name = get_repo_name(repo_url)
    repo_commit_hash = get_commit_hash(repo_url)

    target_base_name = f"{repo_name}_{repo_commit_hash}"

    target_repo_path = target_dir / target_base_name

    return target_repo_path


def clone_repo(
    repo_url: str, target_dir: str | Path | None = None, skip_if_exists: bool = True
) -> Path:
    """Clone a Git repository into a target directory.

    Args:
        repo_url (str): The URL of the Git repository.
        target_dir (Optional[Union[str, Path]]): Target directory to store repo under.
            The repo will be written to a subdirectory. Defaults to temp directory.
        skip_if_exists (bool): Skip cloning if the target directory already exists
            and the commit hash matches. Defaults to True.

    Returns:
        Path to the cloned repository.
    """
    target_path = construct_repo_path(repo_url, target_dir)

    if target_path.exists():
        if skip_if_exists:
            repo_url_commit_hash = get_commit_hash(repo_url)
            try:
                target_path_commit_hash = get_commit_hash(target_path)
            except Exception as e:
                logger.warning(
                    f"An error occurred while checking the commit hash of the existing repository: {e}"  # noqa: E501
                    "Removing the existing path and proceeding with cloning into the following path: "  # noqa: E501
                    f"{target_path}"
                )

            else:
                if target_path_commit_hash == repo_url_commit_hash:
                    # If the commit hashes match, return the existing path
                    logger.info(
                        f"Skipping cloning of repository as target path already exists: {target_path}"  # noqa: E501
                    )
                return target_path
        # If the target path exists but the commit hashes do not match, remove the existing path
        remove_path(target_path)
    try:
        # Clone the repository into the target directory
        base_url, branch = get_repo_url_components(repo_url)
        cmd: list[str] = ["git", "clone", base_url, target_path.as_posix(), "--single-branch"]
        if branch:
            cmd.extend(["--branch", branch])
        subprocess.check_call(cmd)
        return target_path
    except subprocess.CalledProcessError as e:
        logger.error(f"An error occurred while trying to clone the repo: {e}")
        raise e



import aws_cdk as cdk
from aibs_informatics_aws_utils.constants.s3 import (
    S3_SCRATCH_KEY_PREFIX,
    S3_SCRATCH_TAGGING_KEY,
    S3_SCRATCH_TAGGING_VALUE,
)
from aws_cdk import aws_s3 as s3


class LifecycleRuleGenerator:
    @classmethod
    def expire_files_with_scratch_tags(
        cls,
        id: str = "expire-files-with-scratch-tags",
        expiration: cdk.Duration = cdk.Duration.days(30),
        tag_key: str = S3_SCRATCH_TAGGING_KEY,
        tag_value: str = S3_SCRATCH_TAGGING_VALUE,
        enabled: bool = True,
        **lifecycle_rule_kwargs,
    ) -> s3.LifecycleRule:
        return s3.LifecycleRule(
            id=id,
            expiration=expiration,
            tag_filters={tag_key: tag_value},
            enabled=enabled,
            **lifecycle_rule_kwargs,
        )

    @classmethod
    def expire_files_under_prefix(
        cls,
        id: str = "expire-files-under-prefix",
        expiration: cdk.Duration = cdk.Duration.days(30),
        prefix: str = S3_SCRATCH_KEY_PREFIX,
        enabled: bool = True,
        **lifecycle_rule_kwargs,
    ) -> s3.LifecycleRule:
        if id is None:
            id = f"expire-files-under-{prefix}"
        return s3.LifecycleRule(
            id=id,
            expiration=expiration,
            prefix=prefix,
            enabled=enabled,
            **lifecycle_rule_kwargs,
        )

    @classmethod
    def use_storage_class_as_default(
        cls,
        id: str = "use-s3-intelligent-tier-as-default",
        storage_class: s3.StorageClass = s3.StorageClass.INTELLIGENT_TIERING,
        transition_after: cdk.Duration = cdk.Duration.days(0),
        enabled: bool = True,
        **lifecycle_rule_kwargs,
    ) -> s3.LifecycleRule:
        return s3.LifecycleRule(
            id=id,
            transitions=[
                s3.Transition(storage_class=storage_class, transition_after=transition_after)
            ],
            enabled=enabled,
            **lifecycle_rule_kwargs,
        )



import logging
import os
from pathlib import Path

import aws_cdk as cdk
import constructs
from aibs_informatics_core.env import EnvBase
from aibs_informatics_core.utils.decorators import cached_property
from aibs_informatics_core.utils.hashing import generate_path_hash
from aws_cdk import aws_ecr_assets, aws_s3_assets
from aws_cdk import aws_lambda as lambda_

from aibs_informatics_cdk_lib.common.git import clone_repo, is_local_repo, is_repo_url
from aibs_informatics_cdk_lib.constructs_.assets.code_asset import (
    GLOBAL_GLOB_EXCLUDES,
    PYTHON_GLOB_EXCLUDES,
    PYTHON_REGEX_EXCLUDES,
    CodeAsset,
)

AIBS_INFORMATICS_AWS_LAMBDA_REPO_ENV_VAR = "AIBS_INFORMATICS_AWS_LAMBDA_REPO"
AIBS_INFORMATICS_AWS_LAMBDA_REPO = "git@github.com:AllenInstitute/aibs-informatics-aws-lambda.git"

logger = logging.getLogger(__name__)


class AssetsMixin:
    @classmethod
    def resolve_repo_path(cls, repo_url: str, repo_path_env_var: str | None) -> Path:
        """Resolves the repo path from the environment or clones the repo from the url

        This method is useful to quickly swapping between locally modified changes and the remote
        repo.

        This should typically be used in the context of defining a code asset for a static name
        (e.g. AIBS_INFORMATICS_AWS_LAMBDA). You can then use the env var option to point to a local
        repo path for development.

        Args:
            repo_url (str): The git repo url. This is required.
                If the repo path is not in the environment, the repo will be cloned from this url.
            repo_path_env_var (Optional[str]): The environment variable that contains the
                repo path or alternative repo url. This is optional.
                This is useful for local development.

        Returns:
            The path to the repo
        """
        if repo_path_env_var and (repo_path := os.getenv(repo_path_env_var)) is not None:
            logger.info(f"Using {repo_path_env_var} from environment")
            if is_local_repo(repo_path):
                return Path(repo_path)
            elif is_repo_url(str(repo_path)):
                return clone_repo(repo_path, skip_if_exists=True)
            else:
                raise ValueError(f"Env variable {repo_path_env_var} is not a valid git repo")
        else:
            return clone_repo(repo_url, skip_if_exists=True)


class AIBSInformaticsCodeAssets(constructs.Construct, AssetsMixin):
    def __init__(
        self,
        scope: constructs.Construct,
        construct_id: str,
        env_base: EnvBase,
        runtime: lambda_.Runtime | None = None,
        aibs_informatics_aws_lambda_repo: str | None = None,
    ) -> None:
        super().__init__(scope, construct_id)
        self.env_base = env_base
        self.runtime = runtime or lambda_.Runtime.PYTHON_3_11
        self.AIBS_INFORMATICS_AWS_LAMBDA_REPO = (
            aibs_informatics_aws_lambda_repo or AIBS_INFORMATICS_AWS_LAMBDA_REPO
        )

    @cached_property
    def AIBS_INFORMATICS_AWS_LAMBDA(self) -> CodeAsset:
        """Returns a NEW code asset for aibs-informatics-aws-lambda

        Returns:
            The code asset
        """

        repo_path = self.resolve_repo_path(
            self.AIBS_INFORMATICS_AWS_LAMBDA_REPO, AIBS_INFORMATICS_AWS_LAMBDA_REPO_ENV_VAR
        )

        asset_hash = generate_path_hash(
            path=str(repo_path.resolve()),
            excludes=PYTHON_REGEX_EXCLUDES,
        )
        logger.info(f"aibs-informatics-aws-lambda asset hash={asset_hash}")
        bundling_image = self.runtime.bundling_image
        host_ssh_dir = str(Path.home() / ".ssh")
        asset_props = aws_s3_assets.AssetProps(
            # CDK bundles lambda assets in a docker container. This causes issues for our local
            # path dependencies. In order to resolve the relative local path dependency,
            # we need to specify the path to the root of the repo.
            path=str(repo_path),
            asset_hash=asset_hash,
            # It is important to exclude files from the git repo, because
            #   1. it effectively makes our caching for assets moot
            #   2. we also don't want to include certain files for size reasons.
            exclude=[
                *PYTHON_GLOB_EXCLUDES,
                "**/cdk.out/",
                "**/scripts/**",
            ],
            bundling=cdk.BundlingOptions(
                image=bundling_image,
                working_directory="/asset-input",
                entrypoint=["/bin/bash", "-c"],
                command=[
                    # This makes the following commands run together as one
                    # WARNING Make sure not to modify {host_ssh_dir} in any way, in this set of commands!  # noqa: E501
                    " && ".join(
                        [
                            "set -x",
                            # Copy in host ssh keys that are needed to clone private git repos
                            f"cp -r {host_ssh_dir} /root/.ssh",
                            # Useful debug if anything goes wrong with github SSH related things
                            "ssh -vT git@github.com || true",
                            # Must make sure that the package is not installing using --editable mode  # noqa: E501
                            "python3 -m pip install --upgrade pip --no-cache",
                            "pip3 install . --no-cache -t /asset-output",
                            # TODO: remove botocore and boto3 from asset output
                            # Must make asset output permissions accessible to lambda
                            "find /asset-output -type d -print0 | xargs -0 chmod 755",
                            "find /asset-output -type f -print0 | xargs -0 chmod 644",
                        ]
                    ),
                ],
                user="root:root",
                volumes=[
                    cdk.DockerVolume(
                        host_path=host_ssh_dir,
                        container_path=host_ssh_dir,
                    ),
                ],
            ),
        )
        return CodeAsset(
            asset_name=os.path.basename(repo_path.resolve()),
            asset_props=asset_props,
            default_runtime=self.runtime,
            environment={
                self.env_base.ENV_BASE_KEY: self.env_base,
            },
        )


class AIBSInformaticsDockerAssets(constructs.Construct, AssetsMixin):
    def __init__(
        self,
        scope: constructs.Construct,
        construct_id: str,
        env_base: EnvBase,
        aibs_informatics_aws_lambda_repo: str | None = None,
    ) -> None:
        super().__init__(scope, construct_id)
        self.env_base = env_base
        self.AIBS_INFORMATICS_AWS_LAMBDA_REPO = (
            aibs_informatics_aws_lambda_repo or AIBS_INFORMATICS_AWS_LAMBDA_REPO
        )

    @cached_property
    def AIBS_INFORMATICS_AWS_LAMBDA(self) -> aws_ecr_assets.DockerImageAsset:
        """Returns a NEW docker asset for aibs-informatics-aws-lambda

        Returns:
            aws_ecr_assets.DockerImageAsset: The docker asset
        """
        repo_path = self.resolve_repo_path(
            self.AIBS_INFORMATICS_AWS_LAMBDA_REPO, AIBS_INFORMATICS_AWS_LAMBDA_REPO_ENV_VAR
        )

        return aws_ecr_assets.DockerImageAsset(
            self,
            "aibs-informatics-aws-lambda",
            directory=repo_path.as_posix(),
            build_ssh="default",
            platform=aws_ecr_assets.Platform.LINUX_AMD64,
            asset_name="aibs-informatics-aws-lambda",
            file="docker/Dockerfile",
            extra_hash=generate_path_hash(
                path=str(repo_path.resolve()),
                excludes=PYTHON_REGEX_EXCLUDES,
            ),
            exclude=[
                *PYTHON_GLOB_EXCLUDES,
                *GLOBAL_GLOB_EXCLUDES,
            ],
        )


class AIBSInformaticsAssets(constructs.Construct):
    def __init__(
        self,
        scope: constructs.Construct,
        construct_id: str,
        env_base: EnvBase,
        runtime: lambda_.Runtime | None = None,
    ) -> None:
        super().__init__(scope, construct_id)
        self.env_base = env_base

        self.code_assets = AIBSInformaticsCodeAssets(self, "CodeAssets", env_base, runtime=runtime)
        self.docker_assets = AIBSInformaticsDockerAssets(self, "DockerAssets", env_base)



from typing import cast

import constructs
from aibs_informatics_core.env import EnvBase
from aws_cdk import aws_ec2 as ec2
from aws_cdk import aws_route53, aws_route53_targets

from aibs_informatics_cdk_lib.constructs_.base import EnvBaseConstruct


class LimsConnectionConstruct(EnvBaseConstruct):
    """This construct takes as input an AWS ec2.VPC and attaches a "VPC interface endpoint"
    that allows connections to another account/vpc with an on-prem LIMS2 connection.

    vpc_endpoint_service_name should be the DNS name of the service running the LIMS2 connection
    and should look something like: "com.amazonaws.vpce.{region}.vpce-svc-{service_id}"
    """

    def __init__(
        self,
        scope: constructs.Construct,
        id: str | None,
        env_base: EnvBase,
        target_vpc: ec2.Vpc,
        vpc_endpoint_service_name: str,
        **kwargs,
    ):
        super().__init__(scope, id, env_base, **kwargs)

        self.vpc_endpoint_service_name = vpc_endpoint_service_name
        self.target_vpc = target_vpc
        self.add_lims_vpc_endpoint()
        self.add_lims_vpc_endpoint_dns_alias()

    def add_lims_vpc_endpoint(self):
        """Add a VPC endpont to our target_vpc that connects to the LIMS2 endpoint
        service (located in another AWS account/VPC managed by the cloud infra team).

        Useful documentation:
        https://alleninstitute.atlassian.net/wiki/spaces/IT/pages/740360228/Accessing+LIMS2+from+AWS
        https://docs.aws.amazon.com/vpc/latest/privatelink/create-endpoint-service.html#connect-to-endpoint-service
        https://docs.aws.amazon.com/cdk/api/v2/docs/aws-cdk-lib.aws_ec2-readme.html#vpc-endpoints

        Pricing:
        https://aws.amazon.com/privatelink/pricing/ (see: Interface Endpoint pricing section)
        """

        # NOTE: Currently this endpoint only gets associated with 1 subnet because the
        #       endpoint we deploy in our VPC can only be deployed in the same AZs as the
        #       source Service Endpoint. If we would like to have our InterfaceVpcEndpoint
        #       available in multiple AZs, we would need to request cloud infra team to
        #       increase the number of AZs that the source service is deployed to.
        #       See this related question: https://stackoverflow.com/questions/60081850/
        self.lims_vpc_endpoint = ec2.InterfaceVpcEndpoint(
            scope=self,
            id="External LIMS2 Network Load Balancer VPC Endpoint",
            # Obtained from cloud infra team, any changes on their end we also need to update here
            service=ec2.InterfaceVpcEndpointService(name=self.vpc_endpoint_service_name),
            vpc=self.target_vpc,
            open=True,
            lookup_supported_azs=True,
            subnets=ec2.SubnetSelection(one_per_az=True, subnets=self.target_vpc.private_subnets),
        )

        # Need to enable port 80 and 5432 (postgres) for this endpoint
        # https://docs.aws.amazon.com/cdk/api/v2/docs/aws-cdk-lib.aws_ec2-readme.html#allowing-connections
        self.lims_vpc_endpoint.connections.allow_from_any_ipv4(
            port_range=ec2.Port.POSTGRES, description="Postgres port 5432"
        )
        self.lims_vpc_endpoint.connections.allow_from_any_ipv4(
            port_range=ec2.Port.HTTP, description="HTTP port 80"
        )

    def add_lims_vpc_endpoint_dns_alias(self):
        """Add a route53 private hosted zone DNS resolver that will allow us to contact the
        LIMS2 VPC endpoint service using a less unwieldy DNS name.

        Useful documentation:
        https://stackoverflow.com/a/78258885

        Pricing:
        https://aws.amazon.com/route53/pricing/
        """

        # This costs money, but at our usage levels it shouldn't be an issue
        self.private_hosted_zone = aws_route53.PrivateHostedZone(
            scope=self,
            id="Lims2VpcEndpointPrivateHostedZone",
            zone_name="lims2.corp.alleninstitute.org",
            vpc=self.target_vpc,
            comment=(
                "A route53 private hosted zone that contains an Alias that can resolve DNS "
                "queries to `lims2.corp.alleninstitute.org` to the InterfaceVpcEndpoint hosting "
                "the LIMS2 VPC endpoint service"
            ),
        )

        # Use an ARecord (Alias Record) instead of a CRecord because those cost money per query
        alias_record_target = cast(
            aws_route53.IAliasRecordTarget,
            aws_route53_targets.InterfaceVpcEndpointTarget(self.lims_vpc_endpoint),
        )
        self.lims2_vpc_endpoint_arecord = aws_route53.ARecord(
            scope=self,
            id="Lims2VpcEndpointAliasRecord",
            target=aws_route53.RecordTarget.from_alias(alias_target=alias_record_target),
            zone=self.private_hosted_zone,
            delete_existing=True,
        )



from collections.abc import Mapping
from typing import TYPE_CHECKING, Literal

import constructs
from aibs_informatics_aws_utils.batch import (
    build_retry_strategy,
    to_key_value_pairs,
    to_resource_requirements,
)
from aibs_informatics_core.utils.tools.dicttools import convert_key_case
from aibs_informatics_core.utils.tools.strtools import pascalcase
from aws_cdk import aws_stepfunctions as sfn

from aibs_informatics_cdk_lib.constructs_.sfn.utils import convert_reference_paths

if TYPE_CHECKING:
    from mypy_boto3_batch.type_defs import (
        ContainerOverridesTypeDef,
        KeyValuePairTypeDef,
        MountPointTypeDef,
        RegisterJobDefinitionRequestTypeDef,
        VolumeTypeDef,
    )
else:
    ContainerOverridesTypeDef = dict
    KeyValuePairTypeDef = dict
    MountPointTypeDef = dict
    VolumeTypeDef = dict
    RegisterJobDefinitionRequestTypeDef = dict


class BatchOperation:
    @classmethod
    def register_job_definition(
        cls,
        scope: constructs.Construct,
        id: str,
        command: list[str] | str | None,
        image: str,
        job_definition_name: str,
        job_role_arn: str | None = None,
        environment: Mapping[str, str] | str | None = None,
        memory: int | str | None = None,
        vcpus: int | str | None = None,
        gpu: int | str | None = None,
        mount_points: list[MountPointTypeDef] | str | None = None,
        volumes: list[VolumeTypeDef] | str | None = None,
        platform_capabilities: list[Literal["EC2", "FARGATE"]] | str | None = None,
        result_path: str | None = "$",
        output_path: str | None = "$",
    ) -> sfn.Chain:
        """Creates chain to register new job definition

        Following parameters support reference paths:
        - command
        - image
        - job_definition_name
        - environment
        - memory
        - vcpus
        - gpu

        Args:
            scope (constructs.Construct): scope
            id (str): ID prefix
            command (Union[List[str], str]): List of strings or string representing command to run
                Supports reference paths (e.g. "$.foo.bar")
            image (str): image URI or name.
                Supports reference paths (e.g. "$.foo.bar")
            job_definition_name (str): name of job definition.
                Supports reference paths (e.g. "$.foo.bar")
            job_role_arn (Optional[str], optional): Optional job role arn to use for the job.
                Supports reference paths (e.g. "$.foo.bar")
            environment (Optional[Union[Mapping[str, str], str]], optional): Optional environment variables.
                Supports reference paths both as individual values as well as for the entire list of variables.
                However, if a reference path is used for the entire list, the list must be a list of mappings with Name/Value keys".
            memory (Optional[Union[int, str]], optional): Optionally specify memory.
                Supports reference paths (e.g. "$.foo.bar")
            vcpus (Optional[Union[int, str]], optional): Optionally specify . Defaults to None.
            gpu (Optional[Union[int, str]], optional): Optionally specify GPU. Defaults to None.
            mount_points (Optional[List[MountPointTypeDef]], optional): Optionally specify mount points. Defaults to None.
            volumes (Optional[List[VolumeTypeDef]], optional): Optionally specify volumes. Defaults to None.

        Returns:
            sfn.Chain: chain to register job definition
        """  # noqa: E501

        job_definition_name = sfn.JsonPath.format(
            "{}-{}", job_definition_name, sfn.JsonPath.uuid()
        )
        environment_pairs: list[KeyValuePairTypeDef] | str | None
        if not isinstance(environment, str):
            environment_pairs = to_key_value_pairs(dict(environment or {}))
        else:
            environment_pairs = environment

        request: RegisterJobDefinitionRequestTypeDef = {
            "jobDefinitionName": job_definition_name,
            "type": "container",
            "containerProperties": {
                "image": image,
                "command": command,
                "environment": environment_pairs,
                "resourceRequirements": to_resource_requirements(gpu, memory, vcpus),  # type: ignore # must be string
                "mountPoints": mount_points,
                "volumes": volumes,
            },
            "retryStrategy": build_retry_strategy(include_default_evaluate_on_exit_configs=True),
        }  # type: ignore
        if platform_capabilities:
            request["platformCapabilities"] = platform_capabilities  # type: ignore[typeddict-item] # the typing does not understand that this can be a reference path
        if job_role_arn:
            assert "containerProperties" in request  # mollifies mypy
            request["containerProperties"]["jobRoleArn"] = job_role_arn
        parameters = convert_key_case(request, pascalcase)

        start = sfn.Pass(
            scope,
            id + " RegisterJobDefinition Prep",
            parameters=convert_reference_paths(parameters),  # type: ignore  # misundertands type
            result_path=result_path or "$",
        )
        register = sfn.CustomState(
            scope,
            id + " RegisterJobDefinition API Call",
            state_json={
                "Type": "Task",
                "Resource": "arn:aws:states:::aws-sdk:batch:registerJobDefinition",
                "Parameters": {
                    f"{k}.$": f"{result_path if result_path else '$'}.{k}"
                    for k in parameters.keys()
                },
                "ResultSelector": {
                    "JobDefinitionArn.$": "$.JobDefinitionArn",
                    "JobDefinitionName.$": "$.JobDefinitionName",
                    "Revision.$": "$.Revision",
                },
                "ResultPath": result_path,
                "OutputPath": output_path,
                "Retry": [
                    {
                        "ErrorEquals": ["Batch.BatchException"],
                        # Interval at attempt n = IntervalSeconds x BackoffRate ^(n-1)
                        # Total time from first try: 3 + 6 + 12 + 24 + 48 + 96 = 189 seconds
                        "IntervalSeconds": 3,
                        "MaxAttempts": 7,
                        "BackoffRate": 2.0,
                        "JitterStrategy": "FULL",
                    },
                ],
            },
        )
        chain: sfn.Chain | sfn.Pass = start
        if gpu is not None:
            chain = chain.next(
                sfn.Pass(
                    scope,
                    id + " Register Definition Filter Resource Requirements",
                    input_path=f"{result_path or '$'}.ContainerProperties.ResourceRequirements[?(@.Value != 0 && @.Value != '0')]",  # noqa: E501
                    result_path=f"{result_path or '$'}.ContainerProperties.ResourceRequirements",
                )
            )
        return chain.next(register)

    @classmethod
    def submit_job(
        cls,
        scope: constructs.Construct,
        id: str,
        job_name: str,
        job_definition: str,
        job_queue: str,
        parameters: Mapping[str, str] | None = None,
        command: list[str] | str | None = None,
        environment: Mapping[str, str] | str | None = None,
        memory: int | str | None = None,
        vcpus: int | str | None = None,
        gpu: int | str | None = None,
        result_path: str | None = "$",
        output_path: str | None = "$",
    ) -> sfn.Chain:
        job_name = sfn.JsonPath.format(f"{job_name}-{{}}", sfn.JsonPath.uuid())
        environment_pairs: list[KeyValuePairTypeDef] | str | None
        if not isinstance(environment, str):
            environment_pairs = to_key_value_pairs(dict(environment or {}))
        else:
            environment_pairs = environment

        container_overrides: ContainerOverridesTypeDef = {
            "command": command,
            "environment": environment_pairs,  # type: ignore  # mypy can't verify this is the correct type with the current handling of reference paths
            "resourceRequirements": to_resource_requirements(gpu, memory, vcpus),  # type: ignore # must be string
        }  # type: ignore

        request = {
            "JobName": job_name,
            "JobDefinition": job_definition,
            "JobQueue": job_queue,
            "Parameters": parameters or {},
            "ContainerOverrides": container_overrides,
        }
        start = sfn.Pass(
            scope,
            id + " SubmitJob Prep",
            parameters=convert_reference_paths(
                pass_params := convert_key_case(request, pascalcase)  # type: ignore  # misundertands type
            ),
            result_path=result_path or "$",
        )

        submit = sfn.CustomState(
            scope,
            f"{id} SubmitJob API Call",
            state_json={
                "Type": "Task",
                "Resource": "arn:aws:states:::batch:submitJob.sync",
                "Parameters": {
                    f"{k}.$": f"{result_path if result_path else '$'}.{k}"
                    for k in pass_params.keys()
                },
                "ResultSelector": {
                    "JobName.$": "$.JobName",
                    "JobId.$": "$.JobId",
                    "JobArn.$": "$.JobArn",
                },
                "ResultPath": result_path,
                "OutputPath": output_path,
                "Retry": [
                    {
                        "ErrorEquals": ["Batch.BatchException"],
                        # Interval at attempt n = IntervalSeconds x BackoffRate ^(n-1)
                        # Total time from first try: 3 + 6 + 12 + 24 + 48 + 96 = 189 seconds
                        "IntervalSeconds": 3,
                        "MaxAttempts": 7,
                        "BackoffRate": 2.0,
                        "JitterStrategy": "FULL",
                    },
                ],
            },
        )
        chain: sfn.Chain | sfn.Pass = start
        if gpu is not None:
            chain = chain.next(
                sfn.Pass(
                    scope,
                    id + " SubmitJob Filter Resource Requirements",
                    input_path=f"{result_path or '$'}.ContainerOverrides.ResourceRequirements[?(@.Value != 0 && @.Value != '0')]",  # noqa: E501
                    result_path=f"{result_path or '$'}.ContainerOverrides.ResourceRequirements",
                )
            )
        return chain.next(submit)

    @classmethod
    def deregister_job_definition(
        cls,
        scope: constructs.Construct,
        id: str,
        job_definition: str,
        result_path: str | None = "$",
        output_path: str | None = "$",
    ) -> sfn.Chain:
        request = {"jobDefinition": job_definition}
        start = sfn.Pass(
            scope,
            id + " DeregisterJobDefinition Prep",
            parameters=(parameters := convert_key_case(request, pascalcase)),
            result_path=result_path or "$",
        )
        deregister = sfn.CustomState(
            scope,
            f"{id} DeregisterJobDefinition API Call",
            state_json={
                "Type": "Task",
                "Resource": "arn:aws:states:::aws-sdk:batch:deregisterJobDefinition",
                "Parameters": {
                    f"{k}.$": f"{result_path if result_path else '$'}.{k}"
                    for k in parameters.keys()
                },
                "ResultPath": result_path,
                "OutputPath": output_path,
                "Retry": [
                    {
                        "ErrorEquals": ["Batch.BatchException"],
                        # Interval at attempt n = IntervalSeconds x BackoffRate ^(n-1)
                        # Total time from first try: 3 + 6 + 12 + 24 + 48 + 96 = 189 seconds
                        "IntervalSeconds": 3,
                        "MaxAttempts": 7,
                        "BackoffRate": 2.0,
                        "JitterStrategy": "FULL",
                    },
                ],
            },
        )
        return start.next(deregister)



import logging
import os
from collections.abc import Mapping, Sequence
from dataclasses import dataclass
from pathlib import Path

import aws_cdk as cdk
import constructs
from aibs_informatics_core.utils.decorators import cached_property
from aibs_informatics_core.utils.hashing import generate_path_hash
from aws_cdk import aws_lambda as lambda_
from aws_cdk import aws_s3_assets, aws_s3_deployment

logger = logging.getLogger(__name__)


PYTHON_GLOB_EXCLUDES = [
    "**/.git/**",
    "**/*.{egg,egg-info,pyc,pyo}",
    "**/{.egg,.egg-info,.eggs}/**",
    "**/{.venv,__pycache__,build,dist}/**",
]

GLOBAL_GLOB_EXCLUDES = [
    "**/.git/**",
    "**/*.egg",
    "**/.egg-info",
    "**/.pyc",
    "**/.pyo",
    "**/.egg/**",
    "**/.egg-info/**",
    "**/.eggs/**",
    "**/.venv/**",
    "**/__pycache__/**",
    "**/build/**",
    "**/dist/**",
]


PYTHON_REGEX_EXCLUDES = [
    r".*(.git)/.*",
    r".*/(.eggs|.venv|__pycache__|build|dist)/.*",
    r".*.(.egg|.egg-info)/.*",
    r".*.(egg|egg-info|pyc|pyo)/.*",
    r".*/test/.*",
]


@dataclass
class CodeAsset:
    asset_name: str
    asset_props: aws_s3_assets.AssetProps
    default_runtime: lambda_.Runtime
    supported_runtimes: Sequence[lambda_.Runtime] | None = None
    environment: Mapping[str, str] | None = None

    def __post_init__(self):
        if not self.supported_runtimes:
            self.supported_runtimes = [self.default_runtime]

        if self.default_runtime.name not in [_.name for _ in self.supported_runtimes]:
            raise ValueError(
                f"default runtime {self.default_runtime.name} not in list of"
                f"supported runtimes {[_.name for _ in self.supported_runtimes]}"
            )

    @cached_property
    def as_source(self) -> aws_s3_deployment.Source:
        return self.get_source()

    @property
    def as_code(self) -> lambda_.AssetCode:
        return self.get_code()

    def get_environment(self, *overrides: tuple[str, str]) -> Mapping[str, str]:
        environment = {**(self.environment or {})}
        environment.update(overrides)
        return environment

    def get_asset(self, scope: constructs.Construct, id: str) -> aws_s3_assets.Asset:
        return aws_s3_assets.Asset(
            scope,
            id,
            path=self.asset_props.path,
            readers=self.asset_props.readers,
            asset_hash=self.asset_props.asset_hash,
            asset_hash_type=self.asset_props.asset_hash_type,
            bundling=self.asset_props.bundling,
            exclude=self.asset_props.exclude,
            follow_symlinks=self.asset_props.follow_symlinks,
            ignore_mode=self.asset_props.ignore_mode,
        )

    def get_source(self) -> aws_s3_deployment.Source:
        return aws_s3_deployment.Source.asset(
            path=self.asset_props.path,
            readers=self.asset_props.readers,
            asset_hash=self.asset_props.asset_hash,
            asset_hash_type=self.asset_props.asset_hash_type,
            bundling=self.asset_props.bundling,
            exclude=self.asset_props.exclude,
            follow_symlinks=self.asset_props.follow_symlinks,
            ignore_mode=self.asset_props.ignore_mode,
        )  # type: ignore

    def get_source_zip(self, archive_filename: str) -> aws_s3_deployment.Source:
        if not self.asset_props.bundling:
            raise ValueError(
                f"Cannot create nesed zip of source {self.asset_name} with "
                f"asset props = {self.asset_props}, because no bundlings options set!"
            )
        elif not self.asset_props.bundling.command:
            raise ValueError(
                f"Cannot create nesed zip of source {self.asset_name} with "
                f"asset props = {self.asset_props}, because no bundling command set!"
            )
        bundling_command = self.asset_props.bundling.command
        bundling_command[-1] = " && ".join(
            [
                bundling_command[-1],
                "cd /asset-output",
                # Zips everything in the directory, excluding the archive file
                f"zip -r {archive_filename} . -x {archive_filename} -q",
                # deletes everything in the directory, except the archive file
                f"find . ! \\( -name '{archive_filename}' -o -name '.' -o -name '..' \\) -prune -exec rm -rf {{}} +",  # noqa: E501
            ]
        )
        return aws_s3_deployment.Source.asset(
            path=self.asset_props.path,
            readers=self.asset_props.readers,
            asset_hash=self.asset_props.asset_hash,
            asset_hash_type=self.asset_props.asset_hash_type,
            bundling=cdk.BundlingOptions(
                image=self.asset_props.bundling.image,
                command=bundling_command,
                entrypoint=self.asset_props.bundling.entrypoint,
                environment=self.asset_props.bundling.environment,
                local=self.asset_props.bundling.local,
                output_type=cdk.BundlingOutput.NOT_ARCHIVED,
                user=self.asset_props.bundling.user,
                security_opt=self.asset_props.bundling.security_opt,
                volumes=self.asset_props.bundling.volumes,
                working_directory=self.asset_props.bundling.working_directory,
            ),
            follow_symlinks=self.asset_props.follow_symlinks,
            exclude=self.asset_props.exclude,
            ignore_mode=self.asset_props.ignore_mode,
        )  # type: ignore

    def get_code(self) -> lambda_.AssetCode:
        return lambda_.Code.from_asset(
            path=self.asset_props.path,
            readers=self.asset_props.readers,
            asset_hash=self.asset_props.asset_hash,
            asset_hash_type=self.asset_props.asset_hash_type,
            bundling=self.asset_props.bundling,
            exclude=self.asset_props.exclude,
            follow_symlinks=self.asset_props.follow_symlinks,
            ignore_mode=self.asset_props.ignore_mode,
        )

    @classmethod
    def create_py_code_asset(
        cls,
        path: Path,
        context_path: Path | None,
        requirements_file: Path | None = None,
        includes: Sequence[str] | None = None,
        excludes: Sequence[str] | None = None,
        runtime: lambda_.Runtime = lambda_.Runtime.PYTHON_3_11,
        platform: str | None = "linux/amd64",
        environment: Mapping[str, str] | None = None,
        use_uv: bool = False,
    ) -> "CodeAsset":
        """Create and bundle a Python Lambda code asset

        This factory method prepares a Lambda code asset whose source may live in a subdirectory of a
        larger repository. It computes a stable hash (excluding common Python / build artifacts) to
        maximize CDK asset caching, then leverages CDK's Docker bundling to install dependencies into
        /asset-output so they are packaged alongside the function code.

        Dependency installation strategies:
            * Standard pip (default): Installs either the provided requirements file or the local package (.)
                into /asset-output.
            * uv (if use_uv=True): Uses uv for faster, deterministic resolution. If a requirements_file is
                supplied it is installed directly; otherwise uv export produces a frozen requirements file
                which is then installed.

        Private Git dependencies:
            The host's ~/.ssh directory is mounted read‑only into the bundling container so that private
            repositories referenced in dependency specifications (e.g., git+ssh URLs) can be resolved.
            Diagnostic SSH output (ssh -vT git@github.com) is executed to aid debugging but does not fail
            the build.

        File permissions:
            Directory permissions are normalized to 755 and file permissions to 644 to ensure the Lambda
            runtime can read all packaged artifacts.

        Parameters:
            path (Path):
                Path to the Python project or module root you want to package (the leaf containing setup.cfg,
                pyproject.toml, requirements, or the code itself).
            context_path (Optional[Path]):
                Root directory used as the asset's source context (often the monorepo root). If None, defaults
                to path. Everything under this directory (minus excluded patterns) is made available during
                bundling so that relative/local path dependencies can resolve correctly.
            requirements_file (Optional[Path]):
                Path to a requirements file (e.g., requirements.txt). If provided:
                    * pip mode: installs with pip -r <file>
                    * uv mode: installs with uv pip install -r <file>
                If omitted:
                    * pip mode: installs the local project (.)
                    * uv mode: uv export produces a frozen requirements-autogen.txt, then that is installed.
            includes (Optional[Sequence[str]]):
                Additional glob patterns to explicitly include when generating the asset hash. Useful for
                forcing hash sensitivity to files that might otherwise be excluded.
            excludes (Optional[Sequence[str]]):
                Additional glob patterns to exclude (beyond built-in Python / tooling ignores) for hashing
                and packaging. Helps reduce asset size and improve cache hits.
            runtime (aws_cdk.aws_lambda.Runtime):
                Lambda runtime whose bundling image will be used (default: PYTHON_3_11).
            platform (Optional[str]):
                Docker platform string (e.g., linux/amd64) to enforce architecture consistency when building
                on heterogeneous hosts (e.g., Apple Silicon).
            environment (Optional[Mapping[str, str]]):
                Default environment variables to associate with the resulting CodeAsset (may be applied when
                the asset is used in Lambda functions).
            use_uv (bool):
                If True, use uv for dependency resolution and installation; otherwise fall back to pip.

        Returns:
            A new instance encapsulating the bundled Python code and dependencies.
        """  # noqa: E501

        if context_path is None:
            context_path = path

        # This is relative to the repo root copied in asset props below
        package_path = path.relative_to(context_path)
        full_path = context_path / package_path

        asset_hash = generate_path_hash(
            path=str(context_path.resolve()),
            includes=list(includes) if includes else None,
            excludes=[*(excludes or []), *PYTHON_REGEX_EXCLUDES],
        )
        host_ssh_dir = str(Path.home() / ".ssh")

        ssh_setup_commands = [
            # Copy in host ssh keys that are needed to clone private git repos
            # first we make the .ssh directory if it doesn't exist
            "mkdir -p /root/.ssh",
            # then we copy in all files from host ssh dir to container ssh dir
            f"cp -a {host_ssh_dir}/. /root/.ssh/",
            # Useful debug if anything goes wrong with github SSH related things
            "ssh -vT git@github.com || true",
        ]

        uv_setup_commands = [
            # Install uv for this ephemeral bundling container and expose it on PATH
            "curl -LsSf https://astral.sh/uv/install.sh | sh",
            "export PATH=/root/.local/bin:$PATH",
            "uv --version",
        ]
        package_install_commands = []

        if use_uv:
            if requirements_file is not None:
                logging.warning(
                    f"Using uv with requirements file {requirements_file}, "
                    f"make sure that all dependencies are compatible with uv!"
                )
                package_install_commands.append(
                    f"uv pip install -r {requirements_file.as_posix()} --no-cache --target /asset-output",  # noqa: E501
                )
            else:
                package_install_commands += [
                    "uv export --frozen --no-dev --no-editable -o requirements-autogen.txt",
                    "uv pip install --no-sources -r requirements-autogen.txt --no-cache --target /asset-output",  # noqa: E501
                ]
        else:
            package_install_commands += [
                "python3 -m pip install --upgrade pip --no-cache",
                f"python3 -m pip install {'-r ' + requirements_file.as_posix() if requirements_file else '.'} --no-cache --target /asset-output",  # noqa: E501
            ]

        asset_props = aws_s3_assets.AssetProps(
            # CDK bundles lambda assets in a docker container. This causes issues for our local
            # path dependencies. In order to resolve the relative local path dependency,
            # we need to specify the path to the root of the repo.
            path=str(context_path.resolve()),
            asset_hash=asset_hash,
            # It is important to exclude files from the git repo, because
            #   1. it effectively makes our caching for assets moot
            #   2. we also don't want to include certain files for size reasons.
            exclude=[
                *PYTHON_GLOB_EXCLUDES,
                "**/cdk.out/",
            ],
            # ignore_mode=cdk.IgnoreMode.GIT,
            bundling=cdk.BundlingOptions(
                image=runtime.bundling_image,
                working_directory=f"/asset-input/{package_path}",
                entrypoint=["/bin/bash", "-c"],
                command=[
                    # This makes the following commands run together as one
                    # WARNING Make sure not to modify {host_ssh_dir} in any way,
                    # in this set of commands!
                    " && ".join(
                        [
                            "set -x",
                            *ssh_setup_commands,
                            *(uv_setup_commands if use_uv else []),
                            *package_install_commands,
                            # TODO: remove botocore and boto3 from asset output
                            # Must make asset output permissions accessible to lambda
                            "find /asset-output -type d -print0 | xargs -0 chmod 755",
                            "find /asset-output -type f -print0 | xargs -0 chmod 644",
                        ]
                    ),
                ],
                user="root:root",
                volumes=[
                    cdk.DockerVolume(
                        host_path=host_ssh_dir,
                        container_path=host_ssh_dir,
                    ),
                ],
                platform=platform,
            ),
        )

        return CodeAsset(
            asset_name=os.path.basename(full_path.resolve()),
            asset_props=asset_props,
            default_runtime=runtime,
            environment=environment,
        )



"""AWS Batch infrastructure constructs.

This module provides CDK constructs for creating and managing AWS Batch
compute environments, job queues, and related infrastructure.
"""

from collections.abc import Iterable, Mapping, MutableMapping, Sequence
from dataclasses import dataclass
from typing import Literal, cast

import constructs
from aibs_informatics_core.env import EnvBase
from aibs_informatics_core.utils.decorators import cached_property
from aibs_informatics_core.utils.hashing import sha256_hexdigest
from aibs_informatics_core.utils.tools.dicttools import remove_null_values
from aws_cdk import aws_batch as batch
from aws_cdk import aws_ec2 as ec2
from aws_cdk import aws_efs as efs
from aws_cdk import aws_iam as iam
from aws_cdk import aws_s3 as s3

from aibs_informatics_cdk_lib.common.aws.iam_utils import (
    BATCH_READ_ONLY_ACTIONS,
    S3_READ_ONLY_ACCESS_ACTIONS,
    batch_policy_statement,
)
from aibs_informatics_cdk_lib.constructs_.base import EnvBaseConstruct
from aibs_informatics_cdk_lib.constructs_.batch.launch_template import IBatchLaunchTemplateBuilder
from aibs_informatics_cdk_lib.constructs_.batch.types import IBatchEnvironmentDescriptor
from aibs_informatics_cdk_lib.constructs_.efs.file_system import (
    grant_connectable_file_system_access,
    grant_role_file_system_access,
)


class Batch(EnvBaseConstruct):
    """AWS Batch infrastructure construct for creating multiple Batch environments.

    This construct simplifies the creation of Batch Environments.
    It allows for the creation of multiple Batch Environments with different
    configurations and launch templates, but using the same instance role
    and security group.

    Note:
        Instance Roles are created with managed policies commonly used
        by Batch jobs, including access to S3, Lambda, and DynamoDB.

    Defines:
        - Batch Compute Environment (Spot and OnDemand)
        - Instance Role
        - Launch Template
        - Queue(s)
    """

    def __init__(
        self,
        scope: constructs.Construct,
        id: str,
        env_base: EnvBase,
        vpc: ec2.IVpc,
        instance_role_name: str | None = None,
        instance_role_policy_statements: list[iam.PolicyStatement] | None = None,
    ) -> None:
        """Initialize Batch infrastructure construct.

        Creates the shared infrastructure for Batch Environments.
        Has the ability to create multiple Batch Environments with different configurations.

        Args:
            scope (constructs.Construct): The construct scope.
            id (str): The construct ID.
            env_base (EnvBase): Environment base to use for resource naming.
            vpc (ec2.IVpc): VPC to use for Batch infrastructure.
            instance_role_name (Optional[str]): Name for the instance role.
                Defaults to None (auto-generated).
            instance_role_policy_statements (Optional[List[iam.PolicyStatement]]):
                Additional policy statements for the instance role.
                Defaults to None.
        """
        super().__init__(scope, id, env_base)
        self.vpc = vpc

        # ---------------------------------------------------------------------
        # Shared EC2 pieces between the Compute Environments
        #  - instance role/profile
        #  - security group
        #  - launch template
        # ---------------------------------------------------------------------
        self.instance_role = self.create_instance_role(
            role_name=instance_role_name,
            statements=instance_role_policy_statements,
        )
        self.instance_profile = self.create_instance_profile(self.instance_role.role_name)
        self.security_group = self.create_security_group()

        self._batch_environment_mapping: MutableMapping[str, BatchEnvironment] = {}

    @property
    def environments(self) -> list["BatchEnvironment"]:
        """Get all Batch environments sorted by name.

        Returns:
            List of BatchEnvironment instances.
        """
        return sorted(
            self._batch_environment_mapping.values(), key=lambda be: be.descriptor.get_name()
        )

    def create_instance_role(
        self,
        role_name: str | None = None,
        statements: list[iam.PolicyStatement] | None = None,
    ) -> iam.Role:
        """Create an IAM role for Batch EC2 instances.

        Args:
            role_name (Optional[str]): Name for the role. Defaults to None.
            statements (Optional[List[iam.PolicyStatement]]): Additional policy
                statements to attach. Defaults to None.

        Returns:
            The created IAM role.
        """
        instance_role = iam.Role(
            self,
            self.get_child_id(self, "instance-role"),
            role_name=role_name,
            description="Role used by ec2 instance in batch compute environment",
            assumed_by=iam.ServicePrincipal("ec2.amazonaws.com"),  # type: ignore  # Interface not inferred
        )
        managed_policy_names = [
            "service-role/AmazonEC2ContainerServiceforEC2Role",
            "AmazonS3ReadOnlyAccess",
            "AmazonSSMManagedInstanceCore",
            "AmazonElasticFileSystemClientReadWriteAccess",
            "CloudWatchAgentServerPolicy",
        ]
        for mpn in managed_policy_names:
            instance_role.add_managed_policy(iam.ManagedPolicy.from_aws_managed_policy_name(mpn))
        # Used for EBS autoscaling.
        iam.Policy(
            self,
            "ebs-autoscale-policy",
            statements=[
                iam.PolicyStatement(
                    actions=[
                        "ec2:AttachVolume",
                        "ec2:DescribeVolumeStatus",
                        "ec2:DescribeVolumes",
                        "ec2:ModifyInstanceAttribute",
                        "ec2:DescribeVolumeAttribute",
                        "ec2:CreateVolume",
                        "ec2:DeleteVolume",
                        "ec2:CreateTags",
                    ],
                    effect=iam.Effect.ALLOW,
                    resources=["*"],
                )
            ],
            roles=[instance_role],  # type: ignore  # Role is not inferred as IRole
        )

        # Used for S3 / Lambda
        iam.Policy(
            self,
            "s3-ecs-env",
            statements=[
                # allow read access from all buckets
                iam.PolicyStatement(
                    sid="AllReadObjectActions",
                    actions=S3_READ_ONLY_ACCESS_ACTIONS,
                    effect=iam.Effect.ALLOW,
                    resources=[
                        "arn:aws:s3:::*",
                        "arn:aws:s3:::*/*",
                    ],
                ),
                iam.PolicyStatement(
                    sid="AllObjectActions",
                    actions=[
                        "s3:*Object",
                        "s3:GetBucket*",
                        "s3:List*",
                    ],
                    effect=iam.Effect.ALLOW,
                    resources=[
                        f"arn:aws:s3:::{self.env_base}-*",
                        f"arn:aws:s3:::{self.env_base}-*/*",
                    ],
                ),
                iam.PolicyStatement(
                    sid="AllowCallDescribeInstances",
                    actions=["ecs:DescribeContainerInstances"],
                    effect=iam.Effect.ALLOW,
                    resources=["*"],
                ),
                batch_policy_statement(actions=BATCH_READ_ONLY_ACTIONS, env_base=self.env_base),
            ],
            roles=[instance_role],  # type: ignore  # Role is not inferred as IRole
        )

        # Custom policy
        if statements:
            iam.Policy(
                self,
                "custom-policy",
                statements=statements,
                roles=[instance_role],  # type: ignore  # Role is not inferred as IRole
            )

        return instance_role

    def create_instance_profile(self, instance_role_name: str) -> iam.CfnInstanceProfile:
        """Create an instance profile for the Batch instance role.

        Args:
            instance_role_name (str): Name of the IAM role to associate.

        Returns:
            The created instance profile.
        """
        return iam.CfnInstanceProfile(
            self,
            "instance-profile",
            roles=[instance_role_name],
        )

    def create_security_group(self) -> ec2.SecurityGroup:
        """Create a security group for Batch instances.

        Returns:
            The created security group with all outbound and self-ingress allowed.
        """
        security_group = ec2.SecurityGroup(
            self,
            self.get_construct_id("batch", "sg"),
            vpc=self.vpc,
            allow_all_outbound=True,
            description=f"Batch instance security group for {self.env_base}",
        )
        security_group.add_ingress_rule(peer=security_group, connection=ec2.Port.all_traffic())
        return security_group

    def grant_instance_role_permissions(
        self,
        read_write_resources: Iterable[s3.Bucket | efs.FileSystem | efs.IFileSystem] | None = None,
        read_only_resources: Iterable[s3.Bucket | efs.FileSystem | efs.IFileSystem] | None = None,
    ) -> None:
        """Grant the instance role permissions to access resources.

        Args:
            read_write_resources (Optional[Iterable[Union[s3.Bucket, efs.FileSystem, efs.IFileSystem]]]):
                Resources to grant read/write access to.
            read_only_resources (Optional[Iterable[Union[s3.Bucket, efs.FileSystem, efs.IFileSystem]]]):
                Resources to grant read-only access to.

        Raises:
            ValueError: If an unsupported resource type is provided.
        """  # noqa: E501
        for resource in read_write_resources or []:
            if isinstance(resource, s3.Bucket):
                resource.grant_read_write(self.instance_role)
            elif isinstance(resource, efs.FileSystem):
                grant_role_file_system_access(resource, self.instance_role, "rw")  # type: ignore  # Role is not inferred as IRole
            else:
                raise ValueError(f"Unsupported resource type: {resource}")

        for resource in read_only_resources or []:
            if isinstance(resource, s3.Bucket):
                resource.grant_read(self.instance_role)
            elif isinstance(resource, efs.FileSystem):
                grant_role_file_system_access(resource, self.instance_role, "r")  # type: ignore  # Role is not inferred as IRole
            else:
                raise ValueError(f"Unsupported resource type: {resource}")

    def setup_batch_environment(
        self,
        descriptor: IBatchEnvironmentDescriptor,
        config: "BatchEnvironmentConfig",
        launch_template_builder: IBatchLaunchTemplateBuilder | None = None,
    ) -> "BatchEnvironment":
        """Set up a new Batch environment.

        Args:
            descriptor (IBatchEnvironmentDescriptor): Descriptor defining the environment.
            config (BatchEnvironmentConfig): Configuration for the environment.
            launch_template_builder (Optional[IBatchLaunchTemplateBuilder]): Optional
                builder for creating a launch template.

        Returns:
            The created BatchEnvironment construct.
        """
        if launch_template_builder:
            launch_template_builder.grant_instance_role_permissions(self.instance_role)

        # ---------------------------------------------------------------------
        # Batch Environments
        # ---------------------------------------------------------------------
        batch_env = BatchEnvironment(
            self,
            f"{descriptor.get_name()}-batch-env",
            env_base=self.env_base,
            descriptor=descriptor,
            config=config,
            vpc=self.vpc,
            instance_role=self.instance_role,  # type: ignore  # Role is not inferred as IRole
            security_group=self.security_group,
            launch_template_builder=launch_template_builder,
        )
        batch_env.node.add_dependency(self.instance_profile)
        self._batch_environment_mapping[descriptor.get_name()] = batch_env
        return batch_env


DEFAULT_MAXV_CPUS = 10240
DEFAULT_MINV_CPUS = 0


@dataclass
class BatchEnvironmentConfig:
    """Configuration for a Batch compute environment.

    Attributes:
        allocation_strategy (Optional[batch.AllocationStrategy]): The allocation
            strategy for the compute environment.
        instance_types (Optional[List[Union[str, ec2.InstanceType]]]): Instance
            types to use in the compute environment.
        use_public_subnets (bool): Whether to use public subnets.
        use_spot (bool): Whether to use spot instances.
        use_fargate (bool): Whether to use Fargate. Defaults to False.
        maxv_cpus (Optional[int]): Maximum vCPUs for the environment.
            Defaults to 10240.
        minv_cpus (Optional[int]): Minimum vCPUs for the environment.
            Defaults to 0.
    """

    allocation_strategy: batch.AllocationStrategy | None
    instance_types: list[str | ec2.InstanceType] | None
    use_public_subnets: bool
    use_spot: bool
    use_fargate: bool = False
    maxv_cpus: int | None = DEFAULT_MAXV_CPUS
    minv_cpus: int | None = DEFAULT_MINV_CPUS

    def __post_init__(self):
        self.maxv_cpus = min(DEFAULT_MAXV_CPUS, max(1, self.maxv_cpus or DEFAULT_MAXV_CPUS))
        self.minv_cpus = min(1, max(0, self.minv_cpus or DEFAULT_MINV_CPUS))
        if self.use_fargate:
            # https://docs.aws.amazon.com/batch/latest/userguide/fargate.html
            self.minv_cpus = None
            self.instance_types = None

    @property
    def spot_bid_percentage(self) -> int | None:
        """Get the spot bid percentage.

        Returns:
            75 for spot instances (25% discount), None for on-demand.
        """
        spot_bid_percentage = None
        if self.use_spot:
            # We get a 25% discount currently for on-demand compute
            spot_bid_percentage = 75
        return spot_bid_percentage


class BatchEnvironment(EnvBaseConstruct):
    """A single Batch compute environment with its job queue.

    This construct creates a Batch compute environment and associated job queue.
    It supports both Fargate and EC2 compute environments, with customizable
    configurations and launch templates.

    """

    def __init__(
        self,
        scope: constructs.Construct,
        id: str,
        env_base: EnvBase,
        descriptor: IBatchEnvironmentDescriptor,
        config: BatchEnvironmentConfig,
        vpc: ec2.IVpc,
        instance_role: iam.IRole,
        security_group: ec2.SecurityGroup,
        launch_template_builder: IBatchLaunchTemplateBuilder | None = None,
    ):
        """Initialize a BatchEnvironment construct.

        Args:
            scope (constructs.Construct): The construct scope.
            id (str): The construct ID.
            env_base (EnvBase): Environment base for resource naming.
            descriptor (IBatchEnvironmentDescriptor): Environment descriptor.
            config (BatchEnvironmentConfig): Environment configuration.
            vpc (ec2.IVpc): VPC for the environment.
            instance_role (iam.IRole): IAM role for instances.
            security_group (ec2.SecurityGroup): Security group for instances.
            launch_template_builder (Optional[IBatchLaunchTemplateBuilder]):
                Optional launch template builder.
        """
        super().__init__(scope, id, env_base)
        self._descriptor = descriptor
        self.config = config
        self.vpc = vpc
        self.instance_role = instance_role
        self.security_group = security_group
        self.launch_template_builder = launch_template_builder

        # Set up max number of compute environments
        compute_env = batch.OrderedComputeEnvironment(
            compute_environment=self.create_compute_environment(), order=1
        )
        self.compute_environments = [compute_env]
        job_queue = batch.JobQueue(
            self,
            self.job_queue_name,
            job_queue_name=self.job_queue_name,
            compute_environments=self.compute_environments,
        )
        self.job_queue = job_queue

    @property
    def descriptor(self) -> IBatchEnvironmentDescriptor:
        """Get the environment descriptor.

        Returns:
            The batch environment descriptor.
        """
        return self._descriptor

    @property
    def job_queue_name(self) -> str:
        """Get the job queue name.

        Returns:
            The name of the job queue.
        """
        return self.descriptor.get_job_queue_name(self.env_base)

    @property
    def instance_types(self) -> Sequence[ec2.InstanceType] | None:
        """Get the configured instance types.

        Returns:
            List of instance types, or None if not configured.
        """
        if self.config.instance_types:
            return [
                ec2.InstanceType(it) if isinstance(it, str) else it
                for it in self.config.instance_types
            ]
        return None

    @property
    def vpc_subnets(self) -> ec2.SubnetSelection | None:
        """Get the VPC subnet selection.

        Returns:
            SubnetSelection for public subnets if configured, None otherwise.
        """
        vpc_subnets = None
        if self.config.use_public_subnets:
            vpc_subnets = ec2.SubnetSelection(subnets=self.vpc.public_subnets)
        return vpc_subnets

    @cached_property
    def launch_template(self) -> ec2.LaunchTemplate | None:
        """Get or create the launch template.

        Returns:
            The launch template, or None for Fargate environments.
        """
        launch_template = None
        if self.launch_template_builder and not self.config.use_fargate:
            launch_template = self.launch_template_builder.create_launch_template(
                self.descriptor, self.security_group
            )
        return launch_template

    @property
    def launch_template_user_data_hash(self) -> str | None:
        """Get a hash of the launch template user data.

        Returns:
            SHA256 hash of user data, or None if no launch template.
        """
        lt = self.launch_template
        return sha256_hexdigest(lt.user_data.render()) if lt and lt.user_data else None

    @property
    def compute_resource_tags(self) -> Mapping[str, str] | None:
        """Get tags for compute resources.

        Includes a hash of user data to force recreation when launch template changes.

        Returns:
            Dictionary of tags, or None for Fargate environments.
        """
        if not self.config.use_fargate:
            return remove_null_values(
                {
                    "env_base": self.env_base,
                    "batch_queue": self.job_queue_name,
                    "compute_resource_type": self.compute_resource_type,
                    # Get a hash of the launch template data. This is to ensure that when the
                    # launch template changes the ComputeEnvironment will be recreated.
                    # This is required by Batch since any updates to a launch template user data
                    # will not take effect until the Compute Environment itself is destroyed and
                    # recreated.
                    # Related: https://github.com/hashicorp/terraform-provider-aws/issues/15535
                    "launch_template_user_data_hash": self.launch_template_user_data_hash,
                }
            )  # type: ignore  # pyright complains about user data hash when none
        return None

    @property
    def compute_resource_type(self) -> Literal["ON_DEMAND", "SPOT", "FARGATE", "FARGATE_SPOT"]:
        """Get the compute resource type.

        Returns:
            The compute resource type based on configuration.
        """
        if self.config.use_fargate and self.config.use_spot:
            return "FARGATE_SPOT"
        elif self.config.use_fargate and not self.config.use_spot:
            return "FARGATE"
        elif not self.config.use_fargate and self.config.use_spot:
            return "SPOT"
        else:
            return "ON_DEMAND"

    def create_compute_environment(self) -> batch.IComputeEnvironment:
        """Create the compute environment.

        Returns:
            A Fargate or EC2 compute environment based on configuration.
        """
        ce: batch.FargateComputeEnvironment | batch.ManagedEc2EcsComputeEnvironment
        if self.config.use_fargate:
            ce = batch.FargateComputeEnvironment(
                self,
                f"{self.descriptor.get_name()}-ce",
                vpc=self.vpc,
                maxv_cpus=self.config.maxv_cpus,
                spot=self.config.use_spot,
                vpc_subnets=self.vpc_subnets,
            )
        else:
            ce = batch.ManagedEc2EcsComputeEnvironment(
                self,
                f"{self.descriptor.get_name()}-ce",
                vpc=self.vpc,
                allocation_strategy=self.config.allocation_strategy,
                maxv_cpus=self.config.maxv_cpus,
                minv_cpus=self.config.minv_cpus,
                instance_role=self.instance_role,
                launch_template=self.launch_template,
                instance_types=self.instance_types,
                # When instance types are explicitly configured, dont use optimal instance classes
                # because we want to control the instance types via configuration and launch
                # templates. If no instance types are specified, fall back to optimal selection.
                use_optimal_instance_classes=self.instance_types is None,
                spot=self.config.use_spot,
                spot_bid_percentage=self.config.spot_bid_percentage,
                vpc_subnets=self.vpc_subnets,
            )

        if tags := self.compute_resource_tags:
            for tag_key, tag_value in tags.items():
                ce.tags.set_tag(key=tag_key, value=tag_value)
        return ce

    def grant_file_system_access(self, *file_systems: efs.IFileSystem) -> None:
        """Grant compute environments access to EFS file systems.

        Args:
            *file_systems (efs.IFileSystem): Variable number of file systems
                to grant access to.
        """
        for file_system in file_systems:
            for ce in self.job_queue.compute_environments:
                grant_connectable_file_system_access(
                    file_system, cast(ec2.IConnectable, ce.compute_environment)
                )



# Generated from notebooks/ec2-instance-type-selection.ipynb
ON_DEMAND_INSTANCE_TYPES: list[str] = [
    "c5a.large",
    "c5a.xlarge",
    "c5a.2xlarge",
    "c5a.4xlarge",
    "c5a.8xlarge",
    "c5a.12xlarge",
    "c5a.16xlarge",
    "c5a.24xlarge",
    "c6a.large",
    "c6a.xlarge",
    "c6a.2xlarge",
    "c6a.4xlarge",
    "c6a.8xlarge",
    "c6a.12xlarge",
    "c6a.16xlarge",
    "c6a.24xlarge",
    "c6a.32xlarge",
    "c6a.48xlarge",
    "c6a.metal",
    "i3.large",
    "i3.xlarge",
    "i3.2xlarge",
    "i3.4xlarge",
    "i3.8xlarge",
    "i3.16xlarge",
    "inf2.8xlarge",
    "m5.large",
    "m5.xlarge",
    "m5.2xlarge",
    "m5.4xlarge",
    "m5.8xlarge",
    "m5.12xlarge",
    "m5.16xlarge",
    "m5.24xlarge",
    "m5.metal",
    "m5a.large",
    "m5a.xlarge",
    "m5a.2xlarge",
    "m5a.4xlarge",
    "m5a.8xlarge",
    "m5a.12xlarge",
    "m5a.16xlarge",
    "m5a.24xlarge",
    "m5ad.large",
    "m5ad.xlarge",
    "m5ad.2xlarge",
    "m5ad.4xlarge",
    "m5ad.8xlarge",
    "m5ad.12xlarge",
    "m5ad.16xlarge",
    "m5ad.24xlarge",
    "m5d.large",
    "m5d.xlarge",
    "m5d.2xlarge",
    "m5d.4xlarge",
    "m5d.8xlarge",
    "m5d.12xlarge",
    "m5d.16xlarge",
    "m5d.24xlarge",
    "m5d.metal",
    "m6a.large",
    "m6a.xlarge",
    "m6a.2xlarge",
    "m6a.4xlarge",
    "m6a.8xlarge",
    "m6a.12xlarge",
    "m6a.16xlarge",
    "m6a.24xlarge",
    "m6a.32xlarge",
    "m6a.48xlarge",
    "m6a.metal",
    "m6i.large",
    "m6i.xlarge",
    "m6i.2xlarge",
    "m6i.4xlarge",
    "m6i.8xlarge",
    "m6i.12xlarge",
    "m6i.16xlarge",
    "m6i.24xlarge",
    "m6i.32xlarge",
    "m6i.metal",
    "m6id.large",
    "m6id.xlarge",
    "m6id.2xlarge",
    "m6id.4xlarge",
    "m6id.8xlarge",
    "m6id.12xlarge",
    "m6id.16xlarge",
    "m6id.24xlarge",
    "m6id.32xlarge",
    "m6id.metal",
    "m7a.medium",
    "m7a.large",
    "m7a.xlarge",
    "m7a.2xlarge",
    "m7a.4xlarge",
    "m7a.8xlarge",
    "m7a.12xlarge",
    "m7a.16xlarge",
    "m7a.24xlarge",
    "m7a.32xlarge",
    "m7a.48xlarge",
    "m7a.metal-48xl",
    "m7i.large",
    "m7i.xlarge",
    "m7i.2xlarge",
    "m7i.4xlarge",
    "m7i.8xlarge",
    "m7i.12xlarge",
    "m7i.16xlarge",
    "m7i.24xlarge",
    "m7i.48xlarge",
    "m7i.metal-24xl",
    "m7i.metal-48xl",
    "m7i-flex.large",
    "m7i-flex.xlarge",
    "m7i-flex.2xlarge",
    "m7i-flex.4xlarge",
    "m7i-flex.8xlarge",
    "r5.large",
    "r5.xlarge",
    "r5.2xlarge",
    "r5.4xlarge",
    "r5.8xlarge",
    "r5.12xlarge",
    "r5.16xlarge",
    "r5.24xlarge",
    "r5.metal",
    "r5a.large",
    "r5a.xlarge",
    "r5a.2xlarge",
    "r5a.4xlarge",
    "r5a.8xlarge",
    "r5a.12xlarge",
    "r5a.16xlarge",
    "r5a.24xlarge",
    "r5ad.large",
    "r5ad.xlarge",
    "r5ad.2xlarge",
    "r5ad.4xlarge",
    "r5ad.8xlarge",
    "r5ad.12xlarge",
    "r5ad.16xlarge",
    "r5ad.24xlarge",
    "r5d.large",
    "r5d.xlarge",
    "r5d.2xlarge",
    "r5d.4xlarge",
    "r5d.8xlarge",
    "r5d.12xlarge",
    "r5d.16xlarge",
    "r5d.24xlarge",
    "r6a.large",
    "r6a.xlarge",
    "r6a.2xlarge",
    "r6a.4xlarge",
    "r6a.8xlarge",
    "r6a.12xlarge",
    "r6a.16xlarge",
    "r6a.24xlarge",
    "r6a.32xlarge",
    "r6i.large",
    "r6i.xlarge",
    "r6i.2xlarge",
    "r6i.4xlarge",
    "r6i.8xlarge",
    "r6i.12xlarge",
    "r6i.16xlarge",
    "r6i.24xlarge",
    "r6i.32xlarge",
    "r6i.metal",
    "r6id.large",
    "r6id.xlarge",
    "r6id.2xlarge",
    "r6id.4xlarge",
    "r6id.8xlarge",
    "r6id.12xlarge",
    "r6id.16xlarge",
    "r6id.24xlarge",
    "r6id.32xlarge",
    "r6id.metal",
    "r7a.medium",
    "r7a.large",
    "r7a.xlarge",
    "r7a.2xlarge",
    "r7a.4xlarge",
    "r7a.8xlarge",
    "r7a.12xlarge",
    "r7a.16xlarge",
    "r7a.24xlarge",
    "r7a.32xlarge",
    "r7i.large",
    "r7i.xlarge",
    "r7i.2xlarge",
    "r7i.4xlarge",
    "r7i.8xlarge",
    "r7i.12xlarge",
    "r7i.16xlarge",
    "r7i.24xlarge",
    "r7i.metal-24xl",
]


SPOT_INSTANCE_TYPES: list[str] = [
    "c5a.large",
    "c5a.xlarge",
    "c5a.2xlarge",
    "c6a.large",
    "inf2.8xlarge",
    "m5a.large",
    "m5a.xlarge",
    "m5ad.large",
    "m5ad.xlarge",
    "m5ad.2xlarge",
    "m5ad.4xlarge",
    "m5d.large",
    "m5d.xlarge",
    "m5d.8xlarge",
    "m5d.metal",
    "m6a.large",
    "m6a.xlarge",
    "m6a.4xlarge",
    "m6a.metal",
    "m6i.large",
    "m6id.large",
    "m6id.4xlarge",
    "m6id.24xlarge",
    "m7a.medium",
    "m7a.large",
    "m7a.xlarge",
    "m7a.12xlarge",
    "m7a.16xlarge",
    "m7a.32xlarge",
    "m7a.48xlarge",
    "m7i.large",
    "m7i.xlarge",
    "m7i.2xlarge",
    "m7i.12xlarge",
    "m7i.16xlarge",
    "m7i.48xlarge",
    "m7i-flex.large",
    "m7i-flex.xlarge",
    "m7i-flex.2xlarge",
    "r5.large",
    "r5.4xlarge",
    "r5.metal",
    "r5a.large",
    "r5a.xlarge",
    "r5a.2xlarge",
    "r5a.4xlarge",
    "r5a.8xlarge",
    "r5ad.large",
    "r5ad.xlarge",
    "r5ad.2xlarge",
    "r5ad.4xlarge",
    "r5ad.8xlarge",
    "r5ad.12xlarge",
    "r5ad.16xlarge",
    "r5ad.24xlarge",
    "r5d.large",
    "r5d.2xlarge",
    "r5d.4xlarge",
    "r6a.large",
    "r6a.xlarge",
    "r6a.2xlarge",
    "r6a.4xlarge",
    "r6i.large",
    "r6id.large",
    "r6id.xlarge",
    "r6id.2xlarge",
    "r6id.4xlarge",
    "r7a.medium",
    "r7a.16xlarge",
    "r7i.large",
    "r7i.xlarge",
    "r7i.16xlarge",
    "r7i.metal-24xl",
]


TRANSFER_INSTANCE_TYPES: list[str] = [
    "c5.large",
    "c5.xlarge",
    "c5a.large",
    "c5a.xlarge",
    "c5d.large",
    "c5d.xlarge",
    "c5n.large",
    "c6a.large",
    "c6a.xlarge",
    "c6i.large",
    "c6i.xlarge",
    "c6id.large",
    "c6id.xlarge",
    "c7a.medium",
    "c7a.large",
    "c7a.xlarge",
    "c7i.large",
    "c7i.xlarge",
    "inf1.xlarge",
    "m5.large",
    "m5a.large",
    "m5ad.large",
    "m5d.large",
    "m6a.large",
    "m6i.large",
    "m6id.large",
    "m6idn.large",
    "m6in.large",
    "m7a.medium",
    "m7a.large",
    "m7i.large",
    "m7i-flex.large",
    "r7a.medium",
]


LAMBDA_SMALL_INSTANCE_TYPES: list[str] = [
    "c5.large",
    "c5a.large",
    "c5d.large",
    "c6a.large",
    "c6i.large",
    "c6id.large",
    "c7a.medium",
    "c7a.large",
    "c7i.large",
    "m7a.medium",
]


LAMBDA_MEDIUM_INSTANCE_TYPES: list[str] = [
    "c5.xlarge",
    "c5a.xlarge",
    "c5d.xlarge",
    "c5n.large",
    "c6a.xlarge",
    "c6i.xlarge",
    "c6id.xlarge",
    "c7a.xlarge",
    "c7i.xlarge",
    "inf1.xlarge",
    "m5.large",
    "m5a.large",
    "m5ad.large",
    "m5d.large",
    "m6a.large",
    "m6i.large",
    "m6id.large",
    "m6idn.large",
    "m6in.large",
    "m7a.large",
    "m7i.large",
    "m7i-flex.large",
    "r7a.medium",
]


LAMBDA_LARGE_INSTANCE_TYPES: list[str] = [
    "c5.2xlarge",
    "c5a.2xlarge",
    "c5d.2xlarge",
    "c5n.xlarge",
    "c6a.2xlarge",
    "c6i.2xlarge",
    "c6id.2xlarge",
    "c7a.2xlarge",
    "c7i.2xlarge",
    "i3.large",
    "i3en.large",
    "i4i.large",
    "inf1.2xlarge",
    "m5.xlarge",
    "m5a.xlarge",
    "m5ad.xlarge",
    "m5d.xlarge",
    "m6a.xlarge",
    "m6i.xlarge",
    "m6id.xlarge",
    "m6idn.xlarge",
    "m6in.xlarge",
    "m7a.xlarge",
    "m7i.xlarge",
    "m7i-flex.xlarge",
    "r5.large",
    "r5a.large",
    "r5ad.large",
    "r5d.large",
    "r5dn.large",
    "r6a.large",
    "r6i.large",
    "r6id.large",
    "r7a.large",
    "r7i.large",
    "r7iz.large",
]



import inspect
import types
from typing import Any, get_args, get_origin

import aws_cdk as cdk
from aibs_informatics_core.env import EnvBase
from aibs_informatics_core.models.db import DBIndex, DBModel
from aibs_informatics_core.utils.modules import load_type_from_qualified_name
from aws_cdk import aws_dynamodb as dynamodb
from constructs import Construct
from pydantic.fields import FieldInfo

from aibs_informatics_cdk_lib.constructs_.base import EnvBaseConstructMixins


class DynamoDBModelTable(dynamodb.Table):
    def __init__(
        self,
        scope: Construct,
        id: str,
        db_model: type[DBModel],
        db_index: type[DBIndex],
        billing_mode: dynamodb.BillingMode = dynamodb.BillingMode.PAY_PER_REQUEST,
        removal_policy: cdk.RemovalPolicy = cdk.RemovalPolicy.DESTROY,
        stream: dynamodb.StreamViewType | None = None,
        **kwargs,
    ) -> None:
        self.db_model_type = db_model
        self.db_index_type = db_index

        partition_key, sort_key = self.get_partition_and_sort_key()

        super().__init__(
            scope,
            id,
            table_name=self.get_table_name(),
            billing_mode=billing_mode,
            partition_key=partition_key,
            sort_key=sort_key,
            removal_policy=removal_policy,
            stream=stream,
            **kwargs,
        )

        for index in self.db_index_type:
            if index != self.db_index_type.get_default_index():
                self.add_global_secondary_index_from_type(index)

    def get_table_name(self) -> str:
        return self.db_index_type.table_name()

    def get_index_name(self, index: DBIndex) -> str:
        if index.index_name is None:
            raise ValueError("Cannot get index name for default index")
        return index.index_name

    def add_global_secondary_index_from_type(self, index: DBIndex) -> None:
        index_name = self.get_index_name(index)
        partition_key, sort_key = self.get_partition_and_sort_key(index)
        projection_type = dynamodb.ProjectionType.ALL
        non_key_attributes = index.non_key_attributes
        if non_key_attributes is not None:
            projection_type = dynamodb.ProjectionType.INCLUDE

        self.add_global_secondary_index(
            index_name=index_name,
            partition_key=partition_key,
            sort_key=sort_key,
            projection_type=projection_type,
            non_key_attributes=non_key_attributes,
        )

    def get_partition_and_sort_key(
        self, db_index: DBIndex | None = None
    ) -> tuple[dynamodb.Attribute, dynamodb.Attribute | None]:
        db_index = db_index or self.db_index_type.get_default_index()
        db_model_fields = self.db_model_type.model_fields
        partition_key = db_index.key_name
        sort_key = db_index.sort_key_name

        partition_field = db_model_fields.get(partition_key)
        if partition_field is None:
            raise ValueError(
                f"Partition key '{partition_key}' not found in model {self.db_model_type.__name__}"
            )
        partition_attr = dynamodb.Attribute(
            name=partition_key, type=self.get_attribute_type(partition_field)
        )

        sort_attr = None
        if sort_key:
            sort_field = db_model_fields.get(sort_key)
            if sort_field is None:
                raise ValueError(
                    f"Sort key '{sort_key}' not found in model {self.db_model_type.__name__}"
                )
            sort_attr = dynamodb.Attribute(name=sort_key, type=self.get_attribute_type(sort_field))
        return partition_attr, sort_attr

    @classmethod
    def get_attribute_type(cls, field: FieldInfo) -> dynamodb.AttributeType:
        _type: Any = field.annotation
        if _type is None:
            return dynamodb.AttributeType.STRING

        origin = get_origin(_type)
        if origin in (types.UnionType, tuple, list, set, dict) or origin is not None:
            args = [arg for arg in get_args(_type) if arg is not type(None)]
            if len(args) == 1:
                _type = args[0]

        if isinstance(_type, str):
            _type = load_type_from_qualified_name(_type)
        if not inspect.isclass(_type):
            _type = _type.__class__

        if issubclass(_type, int):
            return dynamodb.AttributeType.NUMBER
        elif issubclass(_type, bytes):
            return dynamodb.AttributeType.BINARY
        else:
            return dynamodb.AttributeType.STRING


class EnvBaseDBModelTable(DynamoDBModelTable, EnvBaseConstructMixins):
    def __init__(
        self,
        scope: Construct,
        id: str,
        env_base: EnvBase,
        db_model: type[DBModel],
        db_index: type[DBIndex],
        billing_mode: dynamodb.BillingMode = dynamodb.BillingMode.PAY_PER_REQUEST,
        stream: dynamodb.StreamViewType | None = None,
        **kwargs,
    ) -> None:
        self.env_base = env_base
        super().__init__(
            scope,
            id,
            db_model,
            db_index,
            billing_mode,
            cdk.RemovalPolicy.RETAIN if self.is_test_or_prod else cdk.RemovalPolicy.DESTROY,
            stream,
            **kwargs,
        )

    def get_table_name(self) -> str:
        return self.env_base.get_table_name(self.db_index_type.table_name())

    def get_index_name(self, index: DBIndex) -> str:
        index_name = index.get_index_name(self.env_base)
        if index_name is None:
            raise ValueError("Cannot get index name for default index")
        return index_name



import base64
import functools
import logging
from abc import abstractmethod
from collections.abc import Callable, Mapping, Sequence
from dataclasses import dataclass
from pathlib import Path
from typing import Generic, Optional, TypeVar, cast

import aws_cdk as cdk
import constructs
from aibs_informatics_core.env import EnvBase
from aws_cdk import aws_codepipeline as aws_codepipeline
from aws_cdk import aws_codepipeline_actions, pipelines
from aws_cdk import aws_codestarnotifications as codestarnotifications
from aws_cdk import aws_iam as iam
from aws_cdk import aws_sns as sns
from aws_cdk.aws_codebuild import BuildEnvironment, BuildEnvironmentVariable, BuildSpec

from aibs_informatics_cdk_lib.common.aws.core_utils import build_arn
from aibs_informatics_cdk_lib.common.aws.iam_utils import CODE_BUILD_IAM_POLICY
from aibs_informatics_cdk_lib.project.config import (
    BaseProjectConfig,
    CodePipelineSourceConfig,
    GlobalConfig,
    PipelineConfig,
    StageConfig,
)
from aibs_informatics_cdk_lib.stacks.base import EnvBaseStack

logging.basicConfig(
    level=logging.WARN,
    format="%(asctime)s %(levelname)s %(name)s: %(message)s",
)
logger = logging.getLogger(__name__)


STAGE_CONFIG = TypeVar("STAGE_CONFIG", bound=StageConfig)
GLOBAL_CONFIG = TypeVar("GLOBAL_CONFIG", bound=GlobalConfig)

PIPELINE_STACK = TypeVar("PIPELINE_STACK", bound="BasePipelineStack")


@dataclass
class PipelineStageInfo:
    order: int
    name: str
    pre_steps: list[pipelines.Step] | None = None
    post_steps: list[pipelines.Step] | None = None


def pipeline_stage(
    order: int,
    name: str,
    pre_steps: list[pipelines.Step] | None = None,
    post_steps: list[pipelines.Step] | None = None,
):
    """Method decorator for defining a pipeline stage in a BasePipelineStack subclass.

    you can decorate two types of methods:
    1. A method that returns a cdk.Stage
    2. A method that returns a tuple of pre_steps, cdk.Stage, post_steps
        where pre_steps and post_steps are lists of pipelines.Step objects.

    Example:

    ```python
    class PipelineStack(BasePipelineStack):

        ...

        @pipeline_stage(order=0, name="Source", pre_steps=[...])
        def source_stage(self) -> cdk.Stage:
            return SourceStage(self, self.get_construct_id("source-stage"))

        @pipeline_stage(order=1, name="Build")
        def build_stage(self) -> cdk.Stage:
            return BuildStage(self, self.get_construct_id("build-stage"))

        @pipeline_stage(order=2, name="Deploy")
        def deploy_stage(self) -> Tuple[List[pipelines.Step], cdk.Stage, List[pipelines.Step]]:
            pre_steps = [...]
            post_steps = [...]
            stage = DeployStage(self, self.get_construct_id("deploy-stage"))
            return pre_steps, stage, post_steps

    ```

    Args:
        order (int): Order of the stage. Lower numbers are executed first. E.g. 1, 2, 3, ...
            You can repeat numbers, however, the order will be arbitrary.
        name (str): Name of the stage
        pre_steps (Optional[List[pipelines.Step]], optional): Optional pre steps to add before the stage.
            Defaults to None.
        post_steps (Optional[List[pipelines.Step]], optional): Optional post steps to add after the stage.
            Defaults to None.
    """  # noqa: E501

    def decorator_pipeline_stage(
        func: Callable[[PIPELINE_STACK], cdk.Stage | tuple[cdk.Stage]],
    ) -> Callable[
        [PIPELINE_STACK],
        tuple[Sequence[pipelines.Step] | None, cdk.Stage, Sequence[pipelines.Step] | None],
    ]:
        @functools.wraps(func)
        def wrapper_pipeline_stage(
            *args, **kwargs
        ) -> tuple[Sequence[pipelines.Step] | None, cdk.Stage, Sequence[pipelines.Step] | None]:
            results = func(*args, **kwargs)
            if isinstance(results, cdk.Stage):
                return None, results, None
            assert isinstance(results, tuple) and len(results) == 3
            assert isinstance(results[0], list) or results[0] is None
            assert isinstance(results[1], cdk.Stage)
            assert isinstance(results[2], list) or results[2] is None
            return cast(
                tuple[
                    Optional[Sequence[pipelines.Step]],
                    cdk.Stage,
                    Optional[Sequence[pipelines.Step]],
                ],
                results,
            )

        wrapper_pipeline_stage._pipeline_stage_info = PipelineStageInfo(  # type: ignore[attr-defined]
            order=order, name=name, pre_steps=pre_steps, post_steps=post_steps
        )
        return wrapper_pipeline_stage

    return decorator_pipeline_stage


class BasePipelineStack(EnvBaseStack, Generic[STAGE_CONFIG, GLOBAL_CONFIG]):
    """Defines the CI/CD Pipeline for the an Environment.

    This class is meant to be subclassed to define the pipeline for a specific project.

    You are required to implement the `initialize_pipeline` method which should return
    a `pipelines.CodePipeline` object. You can then add stages to the pipeline by defining
    methods that are decorated with the `@pipeline_stage` decorator.

    Example:

    ```python

    class MyPipelineStack(BasePipelineStack):

        @pipeline_stage(order=1, pre_steps=[...], post_steps=[...])
        def build_stage(self) -> cdk.Stage:
            # Define the steps for the build stage
            build_steps = [
                pipelines.CodeBuildStep(
                    "Build",
                    input=self.get_pipeline_source(self.pipeline_config.source),
                    commands=[
                        "echo 'Building the project'",
                        "npm install",
                        "npm run build",
                    ],
                    role_policy_statements=[
                        self.get_policy_with_secrets(self.pipeline_config.source.oauth_secret_name),
                    ],
                ),
            ]

            # Create the build stage
            build_stage = pipelines.Stage(
                self,
                "BuildStage",
                stage_name="Build",
                actions=build_steps,
            )

            return build_stage


        The following steps are available for pipelines inheriting from this class:
            promotion_stage:    A stage that is added after all other stages. This stage
                                is used to promote the deployment to another environment.
            notifications:      A notification topic that is used to send notifications. You can
                                enable notifications for pipeline failures and successes. This
                                can be configured in the `pipeline_config.notifications` attribute.
    """

    def __init__(
        self,
        scope: constructs.Construct,
        id: str,
        env_base: EnvBase,
        config: BaseProjectConfig[GLOBAL_CONFIG, STAGE_CONFIG],
        **kwargs,
    ) -> None:
        self.project_config = config
        self.stage_config = config.get_stage_config(env_base.env_type)
        self.stage_config.env.label = env_base.env_label
        env = cdk.Environment(
            account=self.stage_config.env.account, region=self.stage_config.env.region
        )
        super().__init__(scope, id, env_base=env_base, env=env, **kwargs)
        self.build_pipeline()

    @abstractmethod
    def initialize_pipeline(self) -> pipelines.CodePipeline:
        raise NotImplementedError("Subclasses must implement this method")

    def build_pipeline(self):
        """Builds the pipeline. This method should be called after the pipeline is initialized.

        This should not be overridden by subclasses unless you know what you are doing.

        This method will:
            1. Initialize the pipeline
            2. Add stages to the pipeline
            3. Add a promotion stage
            4. Build the pipeline
            5. Setup notifications

        """
        # Initialize Pipeline
        self.pipeline = self.initialize_pipeline()

        # Add Stages
        for stage_method in self.get_stage_methods():
            stage_info: PipelineStageInfo = stage_method._pipeline_stage_info  # type: ignore[attr-defined]
            pre_steps, stage, post_steps = stage_method()

            if stage_info.pre_steps is not None:
                pre_steps = [*stage_info.pre_steps, *(pre_steps or [])]
            if stage_info.post_steps is not None:
                post_steps = [*stage_info.post_steps, *(post_steps or [])]

            self.pipeline.add_stage(stage, pre=pre_steps, post=post_steps)

        # Add Promotion Stage
        self.add_promotion_stage(self.pipeline)

        # Build the pipeline
        self.pipeline.build_pipeline()

        # Post Build Setup
        self.setup_notifications(self.pipeline)

    def add_promotion_stage(self, pipeline: pipelines.CodePipeline):
        """Adds a promotion stage to a CodePipeline

        Promotion stages are used to promote the deployment to another environment.
        These promotions are done through github pull requests. This is a major foundation
        for the deployment process.

        The environment promotion definitions are defined in the `global_config.stage_promotions`.
        This is a mapping of source environment types to target environment types.

        The branch that is used for the promotion is defined in `pipeline_config.source.branch`.

        Args:
            pipeline (pipelines.CodePipeline): Code Pipeline
        """
        global_config = self.global_config
        pipeline_config = self.pipeline_config

        # In order to add a CodePipeline Stage without stacks, we must use `add_wave`
        # https://github.com/aws/aws-cdk/issues/15945#issuecomment-895392052
        promote_wave = pipeline.add_wave("Release")

        # POST Steps
        if (source_env_type := self.stage_config.env.env_type) in global_config.stage_promotions:
            promotion_target_env_type = global_config.stage_promotions[source_env_type]
            promotion_target_pipeline_config = self.project_config.get_stage_config(
                promotion_target_env_type
            ).pipeline
            assert promotion_target_pipeline_config is not None
            create_pull_request_step = pipelines.CodeBuildStep(
                "CreateReleasePullRequest",
                input=self.get_pipeline_source(pipeline_config.source),
                # Environment needs to have privelaged access
                build_environment=BuildEnvironment(privileged=True),
                # By default bin/sh is used, so lets set to bash
                # https://docs.aws.amazon.com/codebuild/latest/userguide/build-spec-ref.html#build-spec.shell
                partial_build_spec=BuildSpec.from_object(
                    {
                        "env": {
                            "shell": "bash",
                            "variables": {
                                "CICD_RELEASE_REVIEWER": "AllenInstitute/marmot",
                                "CICD_RELEASE_SOURCE_ENV_TYPE": source_env_type,
                                "CICD_RELEASE_TARGET_ENV_TYPE": promotion_target_env_type,
                                "CICD_RELEASE_TARGET_BRANCH": promotion_target_pipeline_config.source.branch,  # noqa: E501
                            },
                            # https://docs.aws.amazon.com/codebuild/latest/userguide/build-spec-ref.html#build-spec.env.secrets-manager
                            "secrets-manager": {
                                "GITHUB_TOKEN": pipeline_config.source.oauth_secret_name,
                            },
                            "git-credential-helper": "yes",
                        },
                    }
                ),
                install_commands=[
                    # Installing Github CLI (via https://github.com/cli/cli/blob/trunk/docs/install_linux.md)
                    #   1. Resolve Download URL via GH API
                    #   2. Download binary archive
                    #   3. Unarchive and move binary into /usr/local/bin
                    #   4. Verify command is available
                    # Step 1:
                    'GH_CLI_DOWNLOAD_LINK=$(curl -H "Authorization:token $GITHUB_TOKEN" -sSL "https://api.github.com/repos/cli/cli/releases/latest" | jq -r \'.assets[] | select(.name|test(".*_linux_amd64.tar.gz")) | .browser_download_url\')',  # noqa: E501
                    "GH_CLI_TAR_GZ_PATH=$(basename $GH_CLI_DOWNLOAD_LINK)",
                    "GH_CLI_DIR=$(basename $GH_CLI_TAR_GZ_PATH .tar.gz)",
                    # Step 2:
                    'curl -H "Authorization:token $GITHUB_TOKEN" -sSL $GH_CLI_DOWNLOAD_LINK -o $GH_CLI_TAR_GZ_PATH',  # noqa: E501
                    # Step 3:
                    "tar -xf $GH_CLI_TAR_GZ_PATH",
                    "sudo cp $GH_CLI_DIR/bin/gh /usr/local/bin/",
                    # Step 4:
                    "gh --version &> /dev/null",
                ],
                commands=[
                    # Setting up repository WITH git metadata
                    #   Why?
                    #   because Github Version 1 CodePipeline Source does not support
                    #   option for including git metadata. Github Version 2 does this,
                    #   but we cannot use this configuration currently.
                    #   What is going on below?
                    #   1. clone the git repository and work off of that.
                    #   2. Enable caching and store credentials
                    #   3. Checkout branch based on source commit
                    #   4. Run our CI/CD release script
                    "export REPO_DIR=$(mktemp -d)",
                    "cd $REPO_DIR",
                    f"git clone https://${{GITHUB_TOKEN}}@github.com/{pipeline_config.source.repository}.git .",  # noqa: E501
                    # Enables credential caching
                    "git config credential.helper store",
                    # Supposed to force the caching of the credentials
                    "git pull",
                    # Creates a temporary branch using the source commit as its head.
                    #   This ensures that we use the release branch.
                    "git checkout -b $(basename $REPO_DIR) $CODEBUILD_RESOLVED_SOURCE_VERSION",
                    ## Step: Download and run release script
                    # Create a temporary directory and file to store the release script
                    "export RELEASE_SCRIPT_PATH=$(mktemp -d)/cicd-release.sh",
                    "mkdir -p $(dirname $RELEASE_SCRIPT_PATH)",
                    # The release script will not be available to us unless we set up
                    # a virtual environment and install our source package. This is because the
                    # release script is in a dependent package (aibs-informatics-cdk-lib) and
                    # is not included in the source package used as input for this step.
                    # Assuming we want to avoid having to install the package, We have two options:
                    # TODO: Decide which approach is better (prefer 2)
                    #   1. Download the release script from the source repository (using gh cli)
                    #       - This requires the use of the Github CLI
                    #       - This does not couple changes being deployed with the script in repo
                    #       - This is the most direct approach
                    #   2. Base64 encode the release script and decode it on the other side
                    #       - This is a bit more complex
                    #       - This couples changes being deployed with the script in repo
                    (
                        # Download the release script from the source repository (using gh cli)
                        'gh api repos/AllenInstitute/aibs-informatics-cdk-lib/contents/src/aibs_informatics_cdk_lib/cicd/pipeline/scripts/cicd-release.sh --raw -H "Accept: application/vnd.github.v3.raw" > $RELEASE_SCRIPT_PATH'  # noqa: E501
                        if False
                        # Here we are base64 encoding the release script and decoding it on the
                        # other side.
                        # Steps:
                        #   1. Read the release script file
                        #   2. Base64 encode the file
                        #   3. Decode the base64 encoded file and write it to the release script
                        #      path
                        # TODO: i think `importlib.resources.files` is preferred way to go here,
                        #       but it requires specifying the package path. This is a bit more
                        #       difficult to do in this context. So we are using the Path approach.
                        else f"echo {base64.b64encode((Path(__file__).parent / 'scripts' / 'cicd-release.sh').read_text().encode()).decode()} | base64 --decode > $RELEASE_SCRIPT_PATH"  # noqa: E501
                    ),
                    # Run the release script
                    "bash $RELEASE_SCRIPT_PATH",
                ],
                role_policy_statements=[
                    CODE_BUILD_IAM_POLICY,
                    self.get_policy_with_secrets(self.pipeline_config.source.oauth_secret_name),
                ],
            )
            # Add dependencies to all other "post" steps
            if promote_wave.post:
                for post_step in promote_wave.post:
                    create_pull_request_step.add_step_dependency(post_step)

            promote_wave.add_post(create_pull_request_step)

    def setup_notifications(self, pipeline: pipelines.CodePipeline):
        notifications_config = self.pipeline_config.notifications
        if notifications_config.notify_on_any:
            sns_notifications_topic = sns.Topic(
                self,
                self.get_construct_id("sns-notifications"),
                display_name=f"Deployment Pipeline Notifications ({self.env_base})",
                topic_name=f"{self.env_base}-deployment-pipeline-notifications",
            )

            # Pipeline/Stage/Action Failure Notifications
            pipeline.pipeline.notify_on(
                self.get_construct_id("any-failures"),
                target=sns_notifications_topic,  # type: ignore # Topic should match ITopic
                enabled=notifications_config.notify_on_failure,
                events=[
                    aws_codepipeline.PipelineNotificationEvents.PIPELINE_EXECUTION_FAILED,
                ],
                notification_rule_name=f"{self.env_base}-Deployment-Pipeline-Failures",
                detail_type=codestarnotifications.DetailType.FULL,
            )

            # Pipeline Completion Notifications
            pipeline.pipeline.notify_on(
                self.get_construct_id("pipeline-complete"),
                target=sns_notifications_topic,  # type: ignore # Topic should match ITopic
                enabled=notifications_config.notify_on_success,
                events=[
                    aws_codepipeline.PipelineNotificationEvents.PIPELINE_EXECUTION_SUCCEEDED,
                ],
                detail_type=codestarnotifications.DetailType.BASIC,
                notification_rule_name=f"{self.env_base}-Deployment-Pipeline-Success",
            )

    @property
    def project_config(self) -> BaseProjectConfig[GLOBAL_CONFIG, STAGE_CONFIG]:
        return self._project_config

    @project_config.setter
    def project_config(self, value: BaseProjectConfig[GLOBAL_CONFIG, STAGE_CONFIG]):
        self._project_config = value

    @property
    def global_config(self) -> GLOBAL_CONFIG:
        return self.project_config.global_config

    @property
    def stage_config(self) -> STAGE_CONFIG:
        return self._stage_config

    @stage_config.setter
    def stage_config(self, value: STAGE_CONFIG):
        self._stage_config = value

    @property
    def pipeline_config(self) -> PipelineConfig:
        assert self.stage_config.pipeline is not None
        return self.stage_config.pipeline

    @property
    def codebuild_environment_variables(self) -> Mapping[str, BuildEnvironmentVariable]:
        defaults = {
            k: BuildEnvironmentVariable(value=v)
            for k, v in self.stage_config.env.to_env_var_map().items()
        }
        return {
            **self.custom_codebuild_environment_variables,
            **defaults,
        }

    @property
    def custom_codebuild_environment_variables(self) -> Mapping[str, BuildEnvironmentVariable]:
        return {}

    @property
    def source_cache(self) -> dict[str, pipelines.CodePipelineSource]:
        try:
            return self._source_cache
        except AttributeError:
            self.source_cache = {}
        return self._source_cache

    @source_cache.setter
    def source_cache(self, value: dict[str, pipelines.CodePipelineSource]):
        self._source_cache = value

    def get_pipeline_source(
        self, source_config: CodePipelineSourceConfig
    ) -> pipelines.CodePipelineSource:
        """
        Constructs a Github Repo source from a config

        Args:
            source_config (CodePipelineSourceConfig): config

        Returns:
            pipelines.CodePipelineSource:
        """
        # CDK doesnt like when we reconstruct code pipeline source with the same repo name.
        # So we need to cache the results for a given result if config has same repo name.

        if source_config.repository not in self.source_cache:
            if source_config.codestar_connection:
                source = pipelines.CodePipelineSource.connection(
                    repo_string=source_config.repository,
                    branch=source_config.branch,
                    connection_arn=build_arn(
                        service="codestar-connections",
                        resource_type="connection",
                        resource_delim="/",
                        resource_id=source_config.codestar_connection,
                    ),
                    code_build_clone_output=True,
                    trigger_on_push=True,
                )
            elif source_config.oauth_secret_name:
                source = pipelines.CodePipelineSource.git_hub(
                    repo_string=source_config.repository,
                    branch=source_config.branch,
                    authentication=cdk.SecretValue.secrets_manager(
                        secret_id=source_config.oauth_secret_name
                    ),
                    trigger=aws_codepipeline_actions.GitHubTrigger.WEBHOOK,
                )
            else:
                raise ValueError(
                    "Invalid source config. Must have codestar_connection or oauth_secret_name"
                )
            self.source_cache[source_config.repository] = source
        return self.source_cache[source_config.repository]

    def get_stage_methods(
        self,
    ) -> list[Callable[[], tuple[Sequence[pipelines.Step], cdk.Stage, Sequence[pipelines.Step]]]]:
        # Get all methods of the instance
        methods = [
            getattr(self, method_name)
            for method_name in dir(self)
            if callable(getattr(self, method_name))
        ]

        # Filter methods that have the _pipeline_stage_info attribute
        stage_methods = [method for method in methods if hasattr(method, "_pipeline_stage_info")]

        # Sort methods by their order attribute
        stage_methods.sort(key=lambda method: method._pipeline_stage_info.order)  # type: ignore[attr-defined]

        # Return the sorted methods
        return stage_methods

    @staticmethod
    def get_policy_with_secrets(*secret_names: str | None) -> iam.PolicyStatement:
        return iam.PolicyStatement(
            effect=iam.Effect.ALLOW,
            actions=[
                "secretsmanager:GetRandomPassword",
                "secretsmanager:GetResourcePolicy",
                "secretsmanager:GetSecretValue",
                "secretsmanager:DescribeSecret",
                "secretsmanager:ListSecretVersionIds",
            ],
            resources=[
                build_arn(
                    service="secretsmanager",
                    resource_type="secret",
                    resource_delim=":",
                    resource_id=f"{secret_name}-??????",
                )
                for secret_name in secret_names
                if secret_name is not None
            ],
        )



__all__ = [
    "EnvBaseBucket",
    "LifecycleRuleGenerator",
    "grant_bucket_access",
]

from .bucket import EnvBaseBucket, grant_bucket_access
from .lifecycle_rules import LifecycleRuleGenerator
